from typing import Union
from pathlib import Path

from tktkt.util.printing import *
from tktkt.interfaces.tokenisers import *
from tktkt.wrappers.multiplexing import SuccessionalTokeniser
from tktkt.factories.preprocessors import Preprocessor
from tktkt.interfaces.huggingface import TktktToHuggingFace
from tktkt.interfaces.identifiers import NoSpecials, WithSpecials

from ..util.storage import fetchAndCacheDict, DEFAULT_TOKENISER_STEM, makeDownloadPath
from .config import *
from .graph import *


class BTE(TokeniserWithVocabulary[WithSpecials], SuccessionalTokeniser):
    """
    Byte-tuple encoding (BTE): implementation of BPE that can deal with merges of more than 2 parts.
    """

    def __init__(self, preprocessor: Preprocessor,
                 vocab: Vocab[WithSpecials], merges: MergeList,
                 metadata: BTEConfig=None):
        super().__init__(preprocessor=preprocessor, vocab=vocab)
        self._boundary_marker = preprocessor.getBoundaryMarker()

        self.merge_graph: MergeGraph                       = None
        self.merges_starting_with: dict[str, MergeAsTuple] = None  # Will be synchronised with the graph
        self._initialiseGraph(vocab, merges)
        assert self.merge_graph.vocab == self.vocab == vocab

        self._metadata = metadata

    def getName(self) -> str:
        if self._metadata is None:
            # Infer whether it's binary or if there are tuple merges.
            for m in self.merge_graph.merges:
                if len(m.parts) > 2:
                    return "BTE"
            return "BPE"
        else:
            do_prune  = self._metadata.knockout.reference != ReferenceMode.NONE
            do_anneal = self._metadata.annealing.reference != ReferenceMode.NONE
            return "BTE" \
                + ("-knockout-" + self._metadata.knockout.reference.toLetter()) * do_prune \
                + ("-reify" * (self._metadata.reify != ReifyMode.NONE)) \
                + (f"_{self._metadata.iterations}it" if self._metadata.iterations > 0 else "") \
                + (f"_anneal-{self._metadata.annealing.reference.toLetter()}-{self._metadata.annealing.when.toLetter()}") * do_anneal

    def _initialiseGraph(self, vocab: Vocab, mergelist: MergeList, quiet: bool=True):
        self.merge_graph = MergeGraph(vocab, mergelist, quiet=quiet)
        self._syncWithGraph()

    def _syncWithGraph(self):
        """
        Synchronise the class's caching structures with the merge graph, which is the actual knowledge representation of
        the tokeniser's functionality.
        """
        # First, let the merge graph's edits take effect on the identifiers.
        self.merge_graph.vocab.settle()
        assert self.merge_graph.vocab == self.vocab

        # Synchronise merge strings
        padded_merge_rules = self.merge_graph.getPaddedMerges()  # There's no use storing these in one big set/list aside from merges_starting_with, since they're tuples and hence don't have a reference.
        self.merges_starting_with = {t: [] for t in self.merge_graph.vocab}

        for tup in padded_merge_rules:
            head = tup[1][1:-1].split(" ")[0]
            self.merges_starting_with[head].append(tup)  # If this raises a KeyError, something is definitely wrong (merge strings don't match type strings).

    def _initialTokens(self, pretoken: str) -> Tokens:
        """
        BPE requires two special kinds of pretokenisation that aren't really pretokenisation, before tokenising.
            1. It must ensure all spaces have been removed from the input, because these are control characters in the
               merge file and hence they will never partake in any merge. We use them as control characters in the
               algorithm, and hence if pretokenisation didn't get rid of all spaces, we must do so.
               This cannot happen in the preprocessor, because it is a non-invertible mapping AFTER pretokenisation,
               which is invertible.
            2. BPE starts out by splitting up the input into units that can be merged. This is not pretokenisation,
               because these units will interact during tokenisation. The units are usually characters, but they don't
               have to be; Sennrich's repo shows this with an attached end-of-word, e.g. "word" -> "w o r d</w>".
        """
        return self._boundary_marker.atomise(pretoken.replace(" ", ""))

    def _finalTokens(self, tokens: Tokens) -> Tokens:
        """
        :param tokens: All spaces MUST have been removed from the given initial tokens, otherwise really bad things happen.
        """
        buffer = " " + " ".join(tokens) + " "
        while True:
            types = buffer[1:-1].split(" ")
            possible_merges = []
            for t in types:
                for m in self.merges_starting_with.get(t, []):
                    if m[1] in buffer:  # Somehow, 'in' is even faster than slicing at the exact position (buffer[index_in_buffer:index_in_buffer+len(m[1])] == m[1]) which in turn is faster than .startswith ... https://stackoverflow.com/q/31917372/9352077
                        possible_merges.append(m)
                        # print("\t", m[1])  # Note that m[1] is padded with spaces. If not, "a bc d" would allow the merge "a b".

            if not possible_merges:
                break

            best_merge = min(possible_merges)
            buffer = buffer.replace(best_merge[1], best_merge[2])  # TODO: There is an inefficiency bug here that " a b a b ".replace(" a b ", " ab ") == " ab a b " because the middle space cannot match the template twice. The way to solve this would be to join and split on "  " (double space) and have the templates also use double spaces. The problem is that I don't know whether any code relies on getRawMerges()/getPaddedMerges() having exactly one space rather than two between each part.

        return buffer[1:-1].split(" ")

    def _finalTokens_faster(self, sequence_of_nonspaces: Iterable[str]) -> List[str]:  # TODO: Replace _finalTokens by this.
        buffer = " " + " ".join(sequence_of_nonspaces) + " "
        while True:
            tokens = buffer[1:-1].split(" ")
            tokens.pop()  # Slight speedup; the last token in the sequence will never be the initial token of a merge, so it's useless to check merges that start with it.

            possible_merges = []
            for t in set(tokens):  # Slight speedup; don't check for merges of the same type twice. Will be especially important at the start with single-character tokens.
                for m in self.merges_starting_with.get(t, []):
                    if m[1] in buffer:  # Somehow, 'in' is even faster than slicing at the exact position (buffer[index_in_buffer:index_in_buffer+len(m[1])] == m[1]) which in turn is faster than .startswith ... https://stackoverflow.com/q/31917372/9352077
                        possible_merges.append(m)
                        # print("\t", m[1])  # Note that m[1] is padded with spaces. If not, "a bc d" would allow the merge "a b".

            if not possible_merges:
                break

            best_merge = min(possible_merges)
            buffer = buffer.replace(best_merge[1], best_merge[2])

        return buffer[1:-1].split(" ")

    ####################################################################################################################

    @staticmethod
    def from_pretrained_tktkt(checkpoint: Union[str, Path], preprocessor: Preprocessor,
                              specials: WithSpecials=NoSpecials(), unk_id: int=0) -> "BTE[WithSpecials]":
        """
        Load a TkTkT-saved checkpoint (not a HuggingFace-trained checkpoint!) into a TkTkT object.

        Wraps BTE.load(file) to add three features:
            1. You can give a local path as a string, like you would in HuggingFace for local files.
            2. You can give a directory path (as Path or as string), in which case the .json file stem will be imputed.
            3. You can give a string that is NOT an existing file or directory, and it will instead be looked up
               as if it is a checkpoint on the HuggingFace hub.
        """
        # Make sure that you have the tokeniser file locally on disk.
        checkpoint_pathified = Path(checkpoint)
        if checkpoint_pathified.is_dir() or checkpoint_pathified.is_file():
            if checkpoint_pathified.is_dir():
                path = checkpoint_pathified / f"{DEFAULT_TOKENISER_STEM}.json"
                if not checkpoint_pathified.is_file():
                    raise ValueError(f"Couldn't load tokeniser from checkpoint: {checkpoint_pathified.as_posix()} is a directory without recognisable tokeniser file.")
            else:
                path = checkpoint_pathified
        else:  # Comes from a remote location. Could've been cached already though.
            path = makeDownloadPath() / checkpoint.replace("/", "--") / f"{DEFAULT_TOKENISER_STEM}.json"
            if not path.exists():
                # TODO: I wonder how this function fails, actually.
                fetchAndCacheDict(f"https://huggingface.co/{checkpoint}/raw/main/tokenizer.json",
                                  cache_folder=path.parent, stem=DEFAULT_TOKENISER_STEM)
                assert path.exists()

        # Load from disk.
        from .vocabulariser import BPEKnockoutVocabulariser
        types, merges, meta = BPEKnockoutVocabulariser._parseJson(path)
        return BTE(
            preprocessor=preprocessor,
            vocab=Vocab(types, specials=specials, unk_id=unk_id),
            merges=merges,
            metadata=meta
        )

    @staticmethod
    def from_pretrained(checkpoint: Union[str, Path], preprocessor: Preprocessor,
                        specials: WithSpecials=NoSpecials(), unk_id: int=0) -> TktktToHuggingFace:
        """
        Load a TkTkT-saved checkpoint (not a HuggingFace-trained checkpoint!) into a HuggingFace tokeniser.
        If you want to load a HuggingFace-saved checkpoint into a TkTkT tokeniser, use .fromHuggingFace() in any
        of the subclasses of this class.

        Wraps around from_pretrained_tktkt() to give it the HuggingFace interface, which is what you expect from a
        call to from_pretrained().

        Special types are detected automatically. That won't work for all vocabs, but it works for what we need.
        """
        return TktktToHuggingFace(BTE.from_pretrained_tktkt(checkpoint, preprocessor, specials, unk_id))



"""
Goal: BPE-knockout, a post-processing step for BPE where you knock some subwords out of the vocab and rewrite its merge
rules using its two parents. This involves two additional problems solved here:
    1. A BPE tokenisers that can merge triplets, quadruplets ... tuples of any length >= 2.
    2. A way of deciding which types to knock out of the vocabulary. I came up with "blame ratio", see below.

TODO: Because you're already requesting tokenisations and morphological segmentations during blame/amenability anyway,
      you might as well use those to compute a Re, Pr, F1 for free. It would cut the time for doing a diagnostic run in half.
"""
from abc import ABC, abstractmethod
from typing import Any, Self
from collections import Counter
from pathlib import Path
from dataclasses import dataclass, asdict

import re
import json
import dacite
import warnings
from tqdm.auto import tqdm

from modest.interfaces.datasets import ModestDataset

import tktkt  # To have access to tktkt.__version__
from tktkt.util.iterables import cumsum, count, swapped
from tktkt.util.strings import indicesToTokens
from tktkt.util.printing import *
from tktkt.models.bpe.vocabularisation import CacheableBPEArtifacts
from tktkt.factories.artifacts import BPEArtifacts
from tktkt.interfaces.tokenisers import Tokeniser
from tktkt.interfaces.vocabularisers import *

from .. import __version__
from ..util.datahandlers.holdout import Holdout
from ..util.project.config import Pℛ𝒪𝒥ℰ𝒞𝒯, lexiconWeights
from .config import *
from .graph import *
from .tokeniser import BTE


@dataclass
class MergeBlame:
    merge: Merge
    n_bad_applications: int
    n_applications: int

    @property
    def blame_ratio(self) -> float:
        return self.n_bad_applications / self.n_applications if self.n_applications != 0 else 0

    def __lt__(self, other: "MergeBlame"):  # You are worse (i.e. less fit for knockout) when you don't violate boundaries as much.
        return self.blame_ratio < other.blame_ratio


@dataclass
class MergeAmenability:
    merge: str
    n_good_potential_applications: int
    n_potential_applications: int

    @property
    def amenability_ratio(self) -> float:
        return self.n_good_potential_applications / self.n_potential_applications

    def __lt__(self, other: "MergeAmenability"):  # You are worse if you are useful in fewer contexts, or if you are useful in the same amount of contexts, when those contexts are a smaller part of all the contexts you appear in.
        return (self.n_good_potential_applications, self.amenability_ratio) < (other.n_good_potential_applications, other.amenability_ratio)


class IntermediateEvaluator(ABC):
    """
    Interface for evaluating a tokeniser.
    Can be used as an optional argument in tokeniser source code without having to import a testing/visualisation framework in your source.
    """

    @abstractmethod
    def evaluate(self, tokeniser: Tokeniser, holdout: Holdout, experiment_names: list[str]):
        pass


@dataclass
class KnockoutVersions:
    bpe_knockout: str
    tktkt: str


@dataclass
class KnockoutNames:
    tokeniser: str
    dataset: str


@dataclass
class KnockoutIdentification:
    versions: KnockoutVersions
    names: KnockoutNames


@dataclass
class KnockoutRandomisation:
    holdout: float
    seed: int


@dataclass
class KnockoutMetadata:
    identification: KnockoutIdentification
    randomisation: KnockoutRandomisation
    config: BTEConfig
    iterations: list[dict]


class BPEKnockoutArtifacts(BPEArtifacts):
    @abstractmethod
    def getDiagnostics(self) -> KnockoutMetadata:
        pass


class CacheableBPEKnockoutArtifacts(CacheableBPEArtifacts, BPEKnockoutArtifacts):
    def __init__(self, types: list[str], merges: list[tuple[str,...]], metadata: KnockoutMetadata):
        super().__init__(types, merges)
        self._metadata = metadata

    def getDiagnostics(self) -> KnockoutMetadata:
        return self._metadata

    def store(self, cache_path: Path):
        super().store(cache_path)

        with open(cache_path / "metadata.json", "w", encoding="utf-8") as handle:
            json.dump(asdict(self._metadata), handle)

    @classmethod
    def load(cls, cache_path: Path) -> Self:
        bpe_artifacts = CacheableBPEArtifacts.load(cache_path)

        with open(cache_path / "metadata.json", "r", encoding="utf-8") as handle:
            metadata = json.load(handle)

        # Config fields are special, the rest of the metadata is trivial.
        config = metadata["config"]
        config["knockout"]["reference"]  = ReferenceMode(config["knockout"]["reference"])
        config["annealing"]["reference"] = ReferenceMode(config["annealing"]["reference"])
        config["annealing"]["when"]      = AnnealingTime(config["annealing"]["when"])
        config["reify"]                  = ReifyMode(config["reify"])
        metadata["config"] = dacite.from_dict(BTEConfig, config)

        return CacheableBPEKnockoutArtifacts(
            types=bpe_artifacts._types,
            merges=bpe_artifacts._merges,
            metadata=dacite.from_dict(KnockoutMetadata, metadata)
        )

    @classmethod
    def exists(cls, cache_path: Path) -> bool:
        return super().exists(cache_path) and (cache_path / "metadata.json").exists()


EPS = 0.001

# TODO: Should just use the cursor system I have in TkTkT. Might make blame computation much faster.
SPLIT_MARKER = "|"
SPLIT_MARKER_RE = re.compile(re.escape(SPLIT_MARKER))


class BPEKnockoutVocabulariser(SegmentationSupervisedVocabulariser[CacheableBPEKnockoutArtifacts]):

    def __init__(self, initial_tokeniser: BPEArtifacts, config: BTEConfig,
                 holdout: Holdout=None, iteration_evaluator: IntermediateEvaluator=None, quiet: bool=False):
        super().__init__(preprocessor=initial_tokeniser.preprocessorEffective())
        self._config = config
        self._files = initial_tokeniser
        self._print = doPrint(not quiet, hesitate=True)

        if config.iterations < 1 and config.annealing == ReferenceMode.NONE:  # You are literally doing nothing.
            raise ValueError("No iteration nor annealing was requested from the vocabulariser.")

        if config.reify != ReifyMode.NONE and config.knockout == ReferenceMode.NONE:
            raise ValueError(f"Cannot do reification ({config.reify}) without applying knockout first.")

        # Eval
        self._holdout = holdout or Holdout(1.0)
        self._evaluator = iteration_evaluator
        self._diagnostics = []

    def _identifier(self) -> str:
        return "bpe-knockout"

    def _cacheType(self):
        return CacheableBPEKnockoutArtifacts

    def _vocabulariseFromModest(self, reference: ModestDataset) -> CacheableBPEKnockoutArtifacts:
        tk = BTE(
            preprocessor=self._files.preprocessorEffective(),
            vocab=self._files.getVocabulary(),
            merges=self._files.getMerges(),
            metadata=self._config
        )
        self._iterative(tk, reference, iterations=self._config.iterations, evaluator=self._evaluator)
        return self._tokeniserToCache(tk, reference)

    def _iterative(self, tk: BTE, reference: ModestDataset, iterations: int, evaluator: IntermediateEvaluator=None):
        """
        Iterative knockout, with attempts to turn tuple merges back into binary merges (reification) in between.
        This method brings together the three main operations this class performs on BPE graphs:
            - Knockout    (._knockout) to remove types and put a higher-order merge in place;
            - Reification (._reify)    to turn higher-order merges into multiple binary merges;
            - Annealing   (._anneal)   to add merges that would form a token that fills an obvious gap in the vocabulary.
        """
        if iterations > 0 and self._config.knockout.reference == ReferenceMode.NONE:
            raise ValueError(f"Requested {pluralise(iterations, 'iteration')} without knockout configured.")

        self._diagnostics.append({"type": "baseline", "vocab_size": len(tk.merge_graph.vocab)})
        if evaluator:
            evaluator.evaluate(tk, self._holdout, [f"{0} it", "base"])

        # Doing annealing at the start might have some benefit when e.g. two leaf merges will be knocked out, but their
        # combination is a viable merge. In that case, annealing learns the merge, and knockout turns it into a quadruplet.
        if self._config.annealing.when != AnnealingTime.AFTER and self._config.annealing.reference != ReferenceMode.NONE:
            self._anneal(tk, reference)
            if evaluator:
                evaluator.evaluate(tk, self._holdout, [f"{0} it", "+anneal"])

        # Stopping conditions
        END_IF_NO_MORE_DELETIONS = self._config.reify.does_nothing()  # If False, it's possible to just be reifying merges recursively (you reify, do no knockout, then reify again). Note that it's possible to have no knockout in one iteration, but do knockout in the next after adding some novel merges.
        END_IF_NO_MORE_ADDITIONS = False  # If False, can do multiple rounds of knockout without any reification. If True, will cause early stopping when there are no more non-disqualified merges to be suggested, or those that were suggested all need to be created whereas the config demands backwards-compatibility, or if it wasn't, they exist above their triplet.
        DO_KNOCKOUT_IF_NOT_ENDED_ON_IT = True  # Recommendable because the latest additions might be morphologically bad.
        needs_final_knockout = DO_KNOCKOUT_IF_NOT_ENDED_ON_IT and iterations > 0

        all_disqualified_merges: set[str] = set()  # Cache merges that mustn't be retried.
        iteration = 1
        while iteration <= iterations:
            self._print(f"\n=== ITERATION {iteration} ===")

            # --- KNOCKOUT PHASE ---
            removed_merges = self._knockout(tk, reference)
            needs_final_knockout = False
            self._print(f"Knocked out {len(removed_merges)} merges.")

            if END_IF_NO_MORE_DELETIONS and not removed_merges:
                self._print("Early stop: no merges knocked out.")
                break

            if evaluator and removed_merges:
                evaluator.evaluate(tk, self._holdout, [f"{iteration} it", "+knockout"])

            # --- REIFICATION PHASE ---
            if self._config.reify.does_nothing():
                iteration += 1
                continue

            all_disqualified_merges.update(" ".join(m.merge.parts) for m in removed_merges)
            novel_merges = self._reify(tk, all_disqualified_merges)
            needs_final_knockout = len(novel_merges) > 0
            self._print(f"Repaired or reified {len(novel_merges)} merges.")

            if END_IF_NO_MORE_ADDITIONS and not novel_merges:
                self._print("Early stop: no new sub-merges available that weren't knocked out before, nor that exist below their triplet(s).")
                break

            if not removed_merges and not novel_merges:
                self._print("Early stop: tokeniser fully converged (no more deletions, no more additions).")
                self._diagnostics.append({"type": "convergence", "iterations": iteration - 1, "vocab_size": len(tk.merge_graph.vocab)})  # The iteration where you discover that you are identical to previous iteration does not count.
                break

            if evaluator and novel_merges:
                evaluator.evaluate(tk, self._holdout, [f"{iteration} it", "+reify"])
            iteration += 1

        if needs_final_knockout and DO_KNOCKOUT_IF_NOT_ENDED_ON_IT:
            self._print("\n=== FINAL PRUNE ===")
            self._knockout(tk, reference)
            if evaluator:
                evaluator.evaluate(tk, self._holdout, [f"{iteration + 1} it", "+knockout"])

        # Unlike reification, annealing is a linguistically sound post-processing step. It needs no knockout after.
        # You could see it as "filling in the gaps" when you have vocabulary capacity left to e.g. consolidate oversegmented word stems.
        if self._config.annealing.when != AnnealingTime.BEFORE and self._config.annealing.reference != ReferenceMode.NONE:
            self._anneal(tk, reference)
            if evaluator:
                evaluator.evaluate(tk, self._holdout, [f"{iteration + 1} it", "+anneal"])

    ### KNOCKOUT ###

    def _rankOldMergesForKnockout(self, tk: BTE, reference: ModestDataset, blame_tuples_once: bool=False) -> list[MergeBlame]:
        """
        Compares BPE tokenisation to morphological tokenisation, and records the amount of times each BPE merge is used as
        well as the amount of times each merge makes a split disappear that the morphological segmentation mandates.

        Can be repeated before and after knockout; there will always be merges to blame.
        """
        log = doPrint(Pℛ𝒪𝒥ℰ𝒞𝒯.debug_prints)

        segment = self._config.knockout.reference.toMethod()
        weights = lexiconWeights() if self._config.weighted_training else dict()

        merge_lookup = {m.priority: m for m in tk.merge_graph.merges}
        blame        = {m.priority: 0 for m in tk.merge_graph.merges}
        total        = {m.priority: 0 for m in tk.merge_graph.merges}

        for obj in self._holdout(reference.generate(), train=True):
            lemma = obj.word
            weight = weights.get(lemma, 1)
            # log(lemma)

            # Get morphological split
            reference_segmentation = " ".join(segment(obj))
            reference_segmentation = BPEKnockoutVocabulariser._preprocessAlreadySegmentedString(tk, reference_segmentation)

            # Get BPE split and the ID of the merge that caused a space to disappear at each index.
            tokens, merge_ids = BPEKnockoutVocabulariser._prepareAndTokenise_diagnostic(tk, lemma)
            bpe_segmentation = " ".join(tokens)

            # tokens, merge_ids = self._applyMerges_diagnostic(SOW + lemma)
            # bpe_segmentation = " ".join(tokens)[len(SOW):].strip()  # The .strip() is in case the segmentation looks like "Ġ abcd efgh"
            # merge_ids = {k-len(SOW): v for k,v in merge_ids.items() if k >= len(SOW)}  # Shift indices down by the length of the SoW and filter out any entries for it.
            # assert reference_segmentation.replace(" ", "") == bpe_segmentation.replace(" ", "")

            # Get indices with wrongful merges. Unlike compareSplits, we don't use intersection for this.
            # This isn't the only type of error: you can also have too many splits -- a lack of merging -- which can be
            # even worse, e.g. aard schudding -> aard sch udding. We can't (directly) blame any merges for that, though!
            bpe_split_indices = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(      bpe_segmentation).replace("   ", SPLIT_MARKER))}
            ref_split_indices = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(reference_segmentation).replace("   ", SPLIT_MARKER))}
            indices_that_shouldve_never_merged = {index//2 for index in ref_split_indices - bpe_split_indices}  # In an ideal tokeniser, subtracting the BPE split positions should result in an empty set. When it doesn't, BPE is missing split positions, i.e., it merged too many.

            # Blame the merges that caused these indices to contract.
            # log("\t", reference_segmentation, "->", bpe_segmentation)
            # log("\t", merge_ids)
            merge_ids_seen = set()
            for merge_id in merge_ids.values():
                if blame_tuples_once and merge_id in merge_ids_seen:
                    continue
                merge_ids_seen.add(merge_id)
                total[merge_id] += weight

            merge_ids_seen = set()
            for index in indices_that_shouldve_never_merged:
                merge_id = merge_ids[index]  # This will error if the word contains a space, I think.
                # log("\t", f"Blamed: space after '{reference_segmentation.replace(' ', '')[index]}' merged by", merge_lookup[merge_id])
                if blame_tuples_once and merge_id in merge_ids_seen:
                    continue
                merge_ids_seen.add(merge_id)
                blame[merge_id] += weight

        # Calculate ratios
        # blame_ratios = dict()
        # for merge_id in blame.keys():
        #     blame_ratios[merge_id] = blame[merge_id]/total[merge_id] \
        #                              if total[merge_id] != 0 else 0  # Protect against DBZ.
        # filtered_results = [MergeBlame(merge=merge_lookup[merge_id], n_bad_applications=blame[merge_id], n_applications=total[merge_id]) for merge_id, ratio in blame_ratios.items()
        #                     if ratio >= relative_blame_threshold]
        # filtered_results.sort(reverse=True)

        results = [MergeBlame(merge=merge_lookup[merge_id], n_bad_applications=blame[merge_id], n_applications=total[merge_id]) for merge_id in blame.keys()]
        results.sort(reverse=True)
        return results

    def _knockout(self, tk: BTE, reference: ModestDataset) -> list[MergeBlame]:
        self._print("Knockout...")
        merges_to_remove = [m for m in self._rankOldMergesForKnockout(tk, reference, blame_tuples_once=self._config.knockout.blame_tuples_once)
                            if m.blame_ratio >= self._config.knockout.relative_blame_minimum]
        n_eligible = len(merges_to_remove)
        merges_to_remove = merges_to_remove[:max(0, len(tk.merge_graph.vocab) - self._config.knockout.min_vocab_size)]
        explanations = {e.name: count(filter(lambda m: m.merge.explanation == e, merges_to_remove)) for e in MergeExplanation}  # Because applying knockout changes the MergeExplanation, you have to record diagnostics before actually doing knockout.
        self._removeKnockoutMerges(tk, [m.merge for m in merges_to_remove])
        self._diagnostics.append({
            "type": "knockout",
            "eligible": n_eligible,
            "max_offences": max((merge.n_bad_applications for merge in merges_to_remove), default=0),
            "max_blame":    max((merge.blame_ratio        for merge in merges_to_remove), default=0),
            "min_blame":    min((merge.blame_ratio        for merge in merges_to_remove), default=0),
            "vocab_size": len(tk.merge_graph.vocab),
            "explanations": explanations
        })
        return merges_to_remove  # For diagnostic purposes

    def _removeKnockoutMerges(self, tk: BTE, merges_to_remove: Iterable[Merge]):
        for merge in tqdm(merges_to_remove, desc="PRUNING GRAPH", disable=not self._print.verbose):
            if self._config.reify == ReifyMode.NONE_CASCADE:
                try:  # Cascading removes many types, so chances are that one type removes another type to be removed.
                    tk.merge_graph.cascade(merge.childType(), cleanup=True)
                except:
                    pass
            else:
                tk.merge_graph.knockout(merge.childType())
        tk._syncWithGraph()

    ### ANNEALING ###

    def _anneal(self, tk: BTE, reference: ModestDataset) -> list[MergeAmenability]:
        self._print("Annealing...")
        merges_to_add = [m for m in self._rankNewMergesForAnnealing(tk, reference)
                         if m.amenability_ratio >= self._config.annealing.relative_amenability_minimum and m.n_good_potential_applications >= self._config.annealing.absolute_application_minimum]
        n_eligible = len(merges_to_add)

        # Logic to stay within the given vocab size.
        n_max_added = max(0, self._config.annealing.max_vocab_size - len(tk.merge_graph.vocab))
        merges_to_add = merges_to_add[:n_max_added]
        n_can_be_added = n_max_added - len(merges_to_add)

        # Separate check for missing atoms
        missing_atoms = set()
        i = 0
        while i < len(merges_to_add):
            m = merges_to_add[i]
            for part in m.merge.split(" "):
                if part not in missing_atoms and not tk.hasType(part):
                    missing_atoms.add(part)
                    if n_can_be_added > 0:
                        n_can_be_added -= 1
                    else:
                        merges_to_add.pop()
            i += 1

        # Actually do the addition
        self._addAnnealingMerges(tk, [m.merge for m in merges_to_add], add_missing_atoms=True)
        self._diagnostics.append({
            "type": "anneal",
            "eligible": n_eligible,
            "max_good_applications": max((merge.n_good_potential_applications for merge in merges_to_add), default=0),
            "min_good_applications": min((merge.n_potential_applications for merge in merges_to_add), default=0),
            "max_applications": max((merge.n_potential_applications for merge in merges_to_add), default=0),
            "atoms_added": len(missing_atoms),
            "vocab_size": len(tk.merge_graph.vocab)
        })
        return merges_to_add  # For diagnostic purposes

    def _addAnnealingMerges(self, tk: BTE, merges_to_add: Iterable[str], add_missing_atoms: bool=False):
        for merge_string in tqdm(merges_to_add, desc="ANNEALING GRAPH", disable=not self._print.verbose):
            try:
                m = tk.merge_graph.addArc(merge_string, add_missing_atoms=add_missing_atoms)
                m.explanation = MergeExplanation.ANNEALED
            except:
                warn(f"Failed to add arcs for {merge_string} to BPE graph. Probably a missing atom.")
        tk._syncWithGraph()

    def _rankNewMergesForAnnealing(self, tk: BTE, reference: ModestDataset) -> list[MergeAmenability]:
        """
        Suggest merges which improve morphological splits if applied.
        If these involve tokens (characters) that are not in the vocabulary, those will also be added.

        Also note that this cannot fix false negatives, i.e. forgotten splits. That's what knockout is for. For example,
            adembuis
        is split as
            ade mb uis
        which has two false positives
            ade+mb and mb+uis
        yet by doing either of those merges, you don't get the correct split.
        """
        log = doPrint(Pℛ𝒪𝒥ℰ𝒞𝒯.debug_prints)

        segment = self._config.annealing.reference.toMethod()
        weights = lexiconWeights() if self._config.weighted_training else dict()

        do_fuse_spans = False  # Not sure how you would count "total" for this one, unless in a second pass when you already know all the merge spans.
        amenability_count = Counter()
        total_count       = Counter()

        for obj in self._holdout(reference.generate(), train=True):
            lemma = obj.word
            weight = weights.get(lemma, 1)
            log(lemma)

            # Get morphological split
            reference_segmentation = " ".join(segment(obj))
            reference_segmentation = BPEKnockoutVocabulariser._preprocessAlreadySegmentedString(tk, reference_segmentation)

            # Get BPE split
            tokens = tk.prepareAndTokenise(lemma)
            bpe_segmentation = " ".join(tokens)
            log(bpe_segmentation, "->", reference_segmentation)

            # Get indices with wrongful splits, i.e. indices BPE splits at but the reference doesn't say you split at
            # and hence you need to merge. Unlike compareSplits, we don't use intersection for this.
            bpe_split_indices = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(bpe_segmentation      ).replace("   ", SPLIT_MARKER))}
            ref_split_indices = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(reference_segmentation).replace("   ", SPLIT_MARKER))}
            letter_indices_that_need_to_merge = sorted([index//2 + 1 for index in bpe_split_indices - ref_split_indices])

            log(letter_indices_that_need_to_merge)

            # These are all the letters with a space in front of them that shouldn't have a space. Now we want to convert
            # those spaces to the token pairs that merge to delete them.
            # Let's number all tokens, starting at 0, and map each of the letter-after-space indices to such a token identifier.
            lengths = [len(token) for token in tokens]
            letter_indices_that_could_be_merged = [sum(lengths[:i+1]) for i in range(len(lengths)-1)]
            which_token_precedes_that_index = {index: which for which,index in enumerate(letter_indices_that_could_be_merged)}

            log(which_token_precedes_that_index)

            # Get (inclusive) token index spans for token identifiers that actually need to merge.
            token_spans_to_merge = []
            for index in letter_indices_that_need_to_merge:
                left_token_idx = which_token_precedes_that_index[index]
                token_spans_to_merge.append((left_token_idx,left_token_idx+1))

            # E.g.: if you have "a b c" and it needs to be "abc", you'll have the span for "a b" and "b c".
            #       Hence, one approach is to fuse those together into one big "a b c" merge.
            if do_fuse_spans:  # It might also be possible that a span should count for both the multimerge and the individual pair merges. Not sure.
                i = 1
                while i < len(token_spans_to_merge):
                    if token_spans_to_merge[i-1][1] == token_spans_to_merge[i][0]:
                        token_spans_to_merge[i-1:i+1] = [(token_spans_to_merge[i-1][0], token_spans_to_merge[i][1])]
                    else:
                        i += 1
                # Doesn't work:
                # token_indices = [which_token_precedes_that_index[index] for index in letter_indices_to_merge]
                # mask          = [True] + [token_indices[i] != token_indices[i-1]+1 for i in range(1, len(token_indices))] + [True]
                # boundaries    = [i for i in range(len(mask)) if mask[i]]
                # spans         = list(zip(boundaries[:-1], boundaries[1:]))

            # Counting, finally.
            for span_start,span_end in token_spans_to_merge:
                merge_string = " ".join(tokens[span_start:span_end+1])
                amenability_count[merge_string] += weight
                log("\tAmenable:", merge_string)

            for start_token,end_token in zip(tokens[:-1], tokens[1:]):
                merge_string = start_token + " " + end_token
                total_count[merge_string] += weight
                log("\tTotal:", start_token, end_token)

        log("Found", len(total_count), "token combos of which", len(amenability_count), "mended at least one gap.")

        # Filter and ratio
        amenability_ratios = dict()
        for merge,total in total_count.items():
            amenability_ratios[merge] = amenability_count[merge]/total  # Total can't be zero because it wouldn't be in total otherwise.

        # import pandas as pd
        # amenable = {k:v for k,v in amenable.items() if k in amenability_ratios}
        # print(pd.Series(amenable.values()).describe())
        results = [MergeAmenability(merge=merge, n_good_potential_applications=amenability_count[merge], n_potential_applications=total_count[merge]) for merge in amenability_ratios]
        results.sort(reverse=True)
        return results

    ### REIFICATION ###

    def _reify(self, tk: BTE, all_disqualified_merges: set[str]=None) -> set[Merge]:
        """
        In the remaining merges, check which submerges they suggest that aren't disqualified.
        E.g.: there used to be a merge bru + ids. The merge id + s was knocked out.
              - There is now a triplet merge bru + id + s.
              - The submerge id + s is disqualified.
              - The submerge bru + id is available.
        In the code below, when I use "triplet", I mean "thing with submerges". Often it has more than 3 subwords,
        even after just one iteration of knockout, meaning it also suggests more than one submerge.
        """
        if self._config.reify == ReifyMode.NONE:
            return set()

        # Setup
        if all_disqualified_merges is None:
            all_disqualified_merges = set()
        sorted_merges = sorted(tk.merge_graph.merges)
        new_merges    = set()  # Union of (1) triplets that were repaired into different triplets, (2) new binary merges that were applied in at least one triplet, and (3) old binary merges that were applied in at least one other triplet.

        log = doPrint(Pℛ𝒪𝒥ℰ𝒞𝒯.debug_prints)
        diagnostics: dict[str,Any] = {"type": "reify"}

        # Part 1: Find triplets which diverge from the actual tokenisation, and fix them.
        for m in sorted_merges:
            if not self._config.reify.does_fix():  # Skip this whole loop.
                break
            if len(m.parts) <= 2:  # Not a triplet
                continue

            # Find tokenisation up to the triplet.
            # Approach: use the full tokeniser to create some spaces, and add to those spaces the MERGES that happened AFTER the triplet.
            typ = m.childType()
            actual_tokens, causes = BPEKnockoutVocabulariser._tokenise_diagnostic(tk, typ)
            token_lengths = list(cumsum(map(len, actual_tokens)))
            start_of_tokens = {0} \
                            | set(token_lengths[:-1]) \
                            | {space_index+1 for space_index, priority in causes.items() if priority >= m.priority}  # Equivalently, you could construct start_of_tokens subtractively as something like "all possible splits minus the ones taken away by early merges", so set(range(len(typ))) - {space_index+1 for space_index, priority in causes.items() if priority < m.priority}.
            limited_tokens = indicesToTokens(typ, sorted(start_of_tokens))
            assert "".join(limited_tokens) == typ, f"{limited_tokens} vs. {typ}"

            # If the tokens are not as dictated by the merge that is supposed to concatenate them, it needs fixing.
            if limited_tokens != m.parts:
                # self._print(f"Found divergent triplet: expected {m.parts} but got {limited_tokens}")
                m = tk.merge_graph.rewire(typ, limited_tokens)
                m.explanation = MergeExplanation.REPAIRED
                new_merges.add(m)
        diagnostics["triplets_repaired"] = len(new_merges)

        if not self._config.reify.does_link():
            return new_merges

        # Part 2: Find new binary merges inside all triplets.
        suggested_merge_strings: dict[str, set[Merge]] = dict()
        for m in sorted_merges:
            if len(m.parts) <= 2:  # Shortcut; we know that no extra binary merge can be added between the parts because m is literally that merge.
                continue

            for p1, p2 in zip(m.parts[:-1], m.parts[1:]):
                new_binary_merge = p1 + " " + p2
                if new_binary_merge not in all_disqualified_merges:  # Aha, this is a valid merge to try!
                    suggested_merge_strings.setdefault(new_binary_merge, set()).add(m)  # Using a set because the same submerge can appear multiple times in one triplet.
        diagnostics["found_merges"] = len(suggested_merge_strings)

        # ...and try to apply them.
        # Implementation of case 2 and a weak version of case 1.
        #   - Check if submerge doesn't already exist.
        #       - If no: execute case 2 (make the merge with slightly lower priority than the lowest triplet it appears in).
        #       - If yes: check for each triplet it appears in whether it is below it.
        #           - If yes, replace it in the triplet. Not because the triplet results in a different token with or without it, but rather because right now it is preventing the triplet from being applied at all.
        #           - If not, don't do anything (the triplet stays a triplet like vanilla BPE-knockout, and will steal a fraction of all pairs that would go to the merge).
        # Pairs are sorted by how many triplets they appear in.
        diagnostics["applied_merges"] = 0
        diagnostics["created_merges"] = 0
        for merge_string, triplets_this_appears_in in tqdm(sorted(suggested_merge_strings.items(), key=lambda t: len(t[1]), reverse=True), desc="REIFYING SUGGESTIONS", disable=not self._print.verbose):
            parts = merge_string.split()
            subtype = "".join(parts)

            # First of all: it is possible that another submerge took away the opportunity to do this submerge in some of
            # the triplets (e.g. triplet a+b+c+d with suggested merge strings b+c and c+d, and b+c has been carried through
            # in a previous iteration of this loop), so the submerge appears in 0 positions. On the other hand, it might
            # appear not just in one position, but multiple. Both detections are made with the following dict.
            indices_occurs_in_triplet: dict[Merge, list[int]] = dict()
            for triplet in triplets_this_appears_in:
                ignore = False  # A sequence of parts like a+a+a+a only counts as two merges a+a a+a, not three.
                for idx, (p1, p2) in enumerate(zip(triplet.parts[:-1], triplet.parts[1:])):
                    if ignore:
                        ignore = False
                        continue
                    if parts[0] == p1 and parts[1] == p2:  # The merges are all binary, so this is fine.
                        indices_occurs_in_triplet.setdefault(triplet, []).append(idx)
                        ignore = True

            if len(indices_occurs_in_triplet) == 0:
                continue

            # Link or make the new merge.
            if subtype in tk.merge_graph.vocab:  # Link if it's the right pair of tokens. Else do nothing.
                type_is_new = False
                submerge = tk.merge_graph.merges_of[subtype][0]
                if submerge.parts == parts:
                    log(f"Reified merge '{merge_string}' exists.")
                else:
                    log(f"Reified merge '{merge_string}' doesn't exist, but its result is already merged as '{' '.join(submerge.parts)}'. It hence cannot be created.")
                    continue
            else:  # Make new type.
                type_is_new = True
                if self._config.reify.is_backwards_compatible():
                    continue

                log(f"Reified merge '{merge_string}' will be created.")
                submerge = tk.merge_graph.addArc(merge_string)

                # Set the priority to be under the lowest triplet of all triplets that contain it.
                lowest_triplet = min(indices_occurs_in_triplet.keys())
                # TODO: If this triplet is the lowest triplet for another submerge, that submerge will get the same priority...
                #       You probably want a heuristic to order these submerges of the same triplet, such that BPE executes the
                #       best one first. Also, using fractional priorities is a problem because one can be based on another.
                #       Imagine you had an offset of 0.5. Then:
                #           - Merge cd+e with priority 2 is knocked out, so merge ab+cde with priority 3 becomes triplet merge ab+cd+e with priority 3.
                #           - Then ab+cd is chosen as a submerge with priority 2.5, making the triplet a binary merge abcd+e with priority 3.
                #           - Then merge a+b with priority 0 is knocked out and you get a triplet merge a+b+cd with priority 2.5.
                #           - AND NOW, merge b+cd is a submerge whose priority is based on that 2.5.
                submerge.priority = lowest_triplet.priority - EPS

            # In the triplets that are currently blocked by the existence of the submerge (a problem which vanilla BPE-knockout even has), replace the relevant parts by the submerge result.
            merge_had_effect = False  # "was loop not empty"; this can only stay false for merges that (1) already existed and (2) are VERY late.
            for triplet in filter(lambda triplet: submerge < triplet, indices_occurs_in_triplet.keys()):
                merge_had_effect = True

                # 1. Update the tuple inside the triplet's merge node.
                new_parts = []
                i = 0
                while i < len(triplet.parts):
                    if i in indices_occurs_in_triplet[triplet]:
                        new_parts.append(subtype)
                        i += 2
                    else:
                        new_parts.append(triplet.parts[i])
                        i += 1
                triplet.parts = new_parts

                # 2. Link the subtype to the triplet if it isn't already (which would be the case for submerge a+b in a+b+c+ab, but I don't think that happens in practice).
                if triplet not in tk.merge_graph.merges_with[subtype]:
                    tk.merge_graph.merges_with[subtype].append(triplet)
                else:
                    log(triplet, "already known to part", subtype, "which should be impossible")
                    assert False

                # 3. Unlink the parts of this new merge from the triplet (but only if such a part appears nowhere else in the triplet).
                for part in set(parts):  # In case of a merge like a+a.
                    if part not in triplet.parts:
                        tk.merge_graph.merges_with[part].remove(triplet)

            if merge_had_effect:
                new_merges.add(submerge)
                if type_is_new:
                    submerge.explanation = MergeExplanation.REIFIED
                    diagnostics["created_merges"] += 1
                else:
                    submerge.explanation = MergeExplanation.ALREADY_REIFIED
                    diagnostics["applied_merges"] += 1

        tk._syncWithGraph()

        # Diagnostics
        diagnostics["vocab_size"] = len(tk.merge_graph.vocab)
        self._diagnostics.append(diagnostics)
        return new_merges

    ### TOKENISATION METHODS ###

    @staticmethod
    def _preprocessAlreadySegmentedString(tk: BTE, segmentation: str) -> str:
        # TODO: Even this method isn't completely watertight against all preprocessors.
        #       If a preprocessor separates different scripts by a boundary marker, some part of the BPE-knockout code will crash.
        space_preserver_decoded = "🂠"  # Cannot be punctuation or a number since some preprocessors treat that specially. Also can't really be a character in any language.
        space_preserver_encoded, _ = tk._boundary_marker.isolate("".join(tk.preprocessor.do(space_preserver_decoded)))

        segmentation = segmentation.replace(" ", space_preserver_decoded)
        segmentation = "".join(tk.preprocessor.do(segmentation))
        segmentation = segmentation.replace(space_preserver_encoded, " ")
        return segmentation

    @staticmethod
    def _finalTokens_diagnostic(tk: BTE, sequence_of_nonspaces: Iterable[str]) -> tuple[list[str], dict[int, int]]:
        """
        Same as applyMerges, except it returns an extra result (which decreases performance for its computation):
        a map from character index to merge ID. Hence, by calling this version of the function, you can verify which
        merge rule caused the space between two characters to disappear.

        This is even compatible with merges of more than 2 tokens. It's assigned to every missing space after the merge.
        """
        mergepoint_to_mergeid: dict[int,int] = dict()

        buffer = " " + " ".join(sequence_of_nonspaces) + " "
        while True:
            # print(buffer)
            types = buffer[1:-1].split(" ")
            possible_merges: list[MergeAsTuple] = []
            for t in types:
                for m in tk.merges_starting_with.get(t, []):
                    if m[1] in buffer:  # Note that m[1] is padded with spaces. If not, "a bc d" would allow the merge "a b".
                        possible_merges.append(m)
                        # print("\t", m[1])

            if not possible_merges:
                break

            best_merge = min(possible_merges)
            new_buffer = buffer.replace(best_merge[1], best_merge[2])

            # Diagnose which indices where replaced by .replace (can be multiple!)
            split_points_old = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(buffer    ).replace("   ", SPLIT_MARKER))}
            split_points_new = {match.start() for match in SPLIT_MARKER_RE.finditer(" ".join(new_buffer).replace("   ", SPLIT_MARKER))}
            merge_indices    = {index//2 - 1 for index in split_points_old - split_points_new}  # The -1 is because the buffer contains a leading space that isn't in the word.
            # print(best_merge)
            # print(best_merge[1], merge_indices, [word[i] for i in merge_indices])
            for index in merge_indices:
                mergepoint_to_mergeid[index] = best_merge[0]

            buffer = new_buffer

        return buffer[1:-1].split(" "), mergepoint_to_mergeid

    @staticmethod
    def _tokenise_diagnostic(tk: BTE, pretoken: str) -> tuple[list[str], dict[int, int]]:
        return BPEKnockoutVocabulariser._finalTokens_diagnostic(tk, tk._initialTokens(pretoken))

    @staticmethod
    def _prepareAndTokenise_diagnostic(tk: BTE, text: str) -> tuple[list[str], dict[int,int]]:
        """
        Get BPE split and the ID of the merge that caused a space to disappear at each index.
         - You need to pass the lemma to _applyMerges_diagnostic. You cannot just prepend a SoW for this. Here's why:
              - Minor problem: any BPE tokeniser that doesn't have a SoW (and specifically the SoW you use) won't work.
              - Major problem: you'd be segmenting a lemma without converting it to byte-based representation (if
                applicable to your vocab). That means that byte-based BPE will not recognise letters like ë (it is
                used to seeing Ã«) and hence never merge with it, so you can never blame those merges.
         - You need to tackle two problems:
              - Segment strings as if they're in a corpus, which means you need to run the full pretokeniser on
                the input. Technically you want to do exactly what tokeniseAndDecode does,
                which is to say: run preprocessor, tokenise every pretoken, concatenate into one token list, run
                the inverse preprocessor on all tokens, join with spaces, and strip. This is what happens during
                evaluation, where you only need segmentation boundaries.
              - However, you also need to get the index-to-merge diagnosis, whose indices refer to the string as
                seen by the tokeniser rather than the raw lemma string, which is to say:
                      1. originating from a modified version of the string that has both the SoW and the byte characters, and
                      2. if the string has multiple pretokens, each pretoken obviously resets the index to 0.
                The only way around (1) is to pretokenise the reference segmentation to bring its split points
                into the same domain. This will require a hand-crafted pretokeniser, because you can't run a string
                with spaces through the actual preprocessor nor split on those spaces and run each substring through.
                To solve (2), you could concatenate the pretokens, or (because the tokenisers relies on them NOT
                being concatenated) shift all indices up by the length of the previous pretokens.
         - There are two inaccuracies that sneak into the calculation when you compare preprocessed segmentations:
              - You will have even more negatives than there already are, because the positions between special
                bytes (like Ã and «) will never be a split point. Luckily we only care about positives here.
              - The reference will never have a space after SoW or before EoW because we glue it on manually, whilst
                BPE might predict it. For a Latin alphabet, this is very unlikely though, and even if it happens, it
                doesn't affect the results because it isn't a case of BPE merging too much (but rather, too little).
        """
        tokens = []
        merge_ids = dict()

        offset = 0
        for pretoken in tk.preprocessor.do(text):
            partial_tokens, partial_merge_ids = BPEKnockoutVocabulariser._tokenise_diagnostic(tk, pretoken)  # TODO: Technically this treats multi-character start-of-word characters as one character, which will mismatch indices when using this diagnostic to interpret raw string indices. The fix is to increase the first partial_merge_ids by the length of the SoW.

            tokens.extend(partial_tokens)
            merge_ids.update({k + offset: v for k, v in partial_merge_ids.items()})
            offset += len(pretoken)

        return tokens, merge_ids

    #################################################################################################################

    def _tokeniserToCache(self, tk: BTE, reference: ModestDataset) -> CacheableBPEKnockoutArtifacts:
        # Separate alphabet from merged types
        alphabet   = dict(swapped(enumerate(sorted(filter(lambda t: tk.merge_graph.inAlphabet(t), tk.vocab), key=lambda t: (len(t), t)))))
        composites = {m.childType(): m.priority for m in tk.merge_graph.merges}
        if set(tk.vocab.keys()) != set(alphabet.keys()) | set(composites.keys()):
            warnings.warn("While saving, it was discovered that the set of types without a merge plus the set of types formed by the merge list DO NOT make up the vocabulary. This is weird!")

        return CacheableBPEKnockoutArtifacts(
            types=sorted(alphabet.keys(), key=alphabet.get) + sorted(composites.keys(), key=composites.get),  # Alphabet is sorted on length and then alphabetically, composites are sorted by merge priority (even if that means the IDs are out of order, which is the case for RoBERTa!).
            merges=[tuple(m.parts) for m in sorted(tk.merge_graph.merges)],  # Sorted by priority; very important. Isn't necessarily the case in the graph by default.
            metadata=KnockoutMetadata(
                identification=KnockoutIdentification(
                    versions=KnockoutVersions(
                        bpe_knockout=__version__,
                        tktkt=tktkt.__version__
                    ),
                    names=KnockoutNames(
                        tokeniser=tk.getName(),
                        dataset=reference.identifier() + (f"_{int(100*self._holdout.threshold)}-{int(100-100*self._holdout.threshold)}-holdout" if self._holdout is not None and self._holdout.threshold != 1.0 else "")
                    )
                ),
                randomisation=KnockoutRandomisation(
                    holdout=self._holdout.threshold,
                    seed=self._holdout.seed
                ),
                config=self._config,
                iterations=self._diagnostics
            )
        )



from datasets import load_dataset, Dataset, IterableDataset
from tktkt.util.types import NamedIterable
from torch.utils.data import DataLoader
from tokenizers import Regex
import tokenizers.normalizers as tn
import tokenizers.pre_tokenizers as tp

from tktkt.models.word.vocabularisation import CountWords
from tktkt.factories.preprocessors import TraditionalPreprocessor

from .wordfiles import *
from .unicode import punctuation_regex_str


def logger(msg: str):
    print("[" + time.strftime('%H:%M:%S') + "]", msg)


def generateDataloader_Oscar(langtag: str, sentence_preprocessor: Callable[[str],str],
                             size_limit: int=None, shuffle: bool=False, streamed: bool=True) -> Tuple[DataLoader, int]:
    """
    Note that the DataLoader is an iteraBLE, not an iteraTOR. It can be iterated over multiple times.
    """
    if not streamed and langtag != "en":
        logger("Loading dataset... (takes about 5 minutes for NL and 10 minutes for DE)")
        data: Dataset = load_dataset(path="oscar", name="unshuffled_deduplicated_" + langtag, split="train")
        logger("Finished loading.")

        size = len(data)
        if size_limit is not None and size > size_limit:
            size = size_limit
            # Shuffling has pros and cons:
            #   Pro: you avoid one topic/document dominating the dataset. Especially useful for small subsets (or large source documents).
            #   Con: your hard drive isn't doing sequential reads, which has at least two downsides:
            #           (1) Random disk access is known to be much slower. In practice: 30-hour ETA vs. 2-hour ETA.
            #           (2) It is strenuous for the device. I can hear my drive crackling very loudly after shuffling.
            if shuffle:
                data = data.shuffle(seed=0)
                indices = range(size_limit)
                # 10s of millions of indices is >100 MiB of memory. HuggingFace .shuffle() caches indices for a reason.
                # import numpy.random as npr
                # rng = npr.default_rng(seed=0)
                # indices = rng.choice(len(data), size=size_limit, replace=False)
            else:
                indices = range(size_limit)

            data = data.select(indices)

    else:  # OSCAR-en is literally used by HuggingFace as example of dataset that is too gigantic to download (1.2 TiB...) https://huggingface.co/docs/datasets/stream
        logger(f"Streaming OSCAR {langtag}.")
        data: IterableDataset = load_dataset(path='oscar-corpus/oscar', name="unshuffled_deduplicated_en",
                                             split='train', streaming=True)
        size = data.info.splits["train"].num_examples  # bruh who invented this interface
        if size_limit is not None and size > size_limit:
            size = size_limit
            data = data.take(size_limit)

    # data = data.remove_columns(["id"])
    def dictionaryProcessor(batched_example):
        """
        If you iterate over 'data', you get dictionaries.
        We want iteration to cause raw strings, which is what the BPE interface requires.
        This processor catches a dictionary of 1 sentence and converts it to a raw string.
        """
        return sentence_preprocessor([example["text"] for example in batched_example][0])

    return DataLoader(data, shuffle=False, collate_fn=dictionaryProcessor), size


def dataloaderToCounts(dataloader: DataLoader, output_stem: str):
    counter = CountWords(
        word_extractor=TraditionalPreprocessor(),
        frequency_minimum=10,
        sort_before_write=True,
        cache_config=CountWords.CacheConfig(
            checkpoint_every_examples=1_000_000,
            flush_if_keys_exceed=1_000_000,
            drop_if_multiple_exceeded=3,
            delete_cache_after=True
        )
    )
    return counter.vocabulariseFromStringIterable(NamedIterable(dataloader, name=output_stem))


def punctuationPretokeniserExceptHyphens():
    punctuation_regex_str_no_hyphenish = punctuation_regex_str.replace("\\-", "").replace("–", "").replace("_", "")
    pretokeniser = tp.Split(pattern=Regex(punctuation_regex_str_no_hyphenish),
                            behavior="isolated")
    normalizer = tn.NFKC()  # Turn weird letters into normal letters, and leave accents on top of their letters. TODO: For Dutch, you want to remove all ´ and ` but keep all ¨ and ^. There is no normaliser for that.

    def wordSeparator(s: str) -> str:
        return " ".join([w.strip() for w, _ in pretokeniser.pre_tokenize_str(normalizer.normalize_str(s))])

    return wordSeparator



from typing import TextIO, Tuple, Optional, Callable, Dict
from collections import Counter
from pathlib import Path

import gc
from tqdm.auto import tqdm

from .unicode import *
from tktkt.util.timing import *
from tktkt.util.printing import *

from ..storage import makeDownloadPath


def iterateTxt(open_file_handle: TextIO, verbose=False):
    """
    Here's how this function works:
        - Python recognises that a 'yield' is used and not a 'return'. Hence, when you call the function, all that is
          returned is a generator object that has stored the arguments to the function and nothing else.
        - When you iterate over the result of the function call, the first iteration will run until the yield and
          return its result. The next iteration, it will continue running past that yield until it encounters it again.
    """
    open_file_handle.seek(0)
    if verbose:
        # Count total lines
        total_lines = 0
        for _ in open_file_handle:
            total_lines += 1
        open_file_handle.seek(0)

        # Now generate each line whilst updating a progress bar.
        for line in tqdm(open_file_handle, total=total_lines, desc=Path(open_file_handle.name).name, smoothing=0.05):
            yield line.rstrip()
    else:
        for line in open_file_handle:
            yield line.rstrip()


def iterateWordsFile(open_file_handle: TextIO, sep=" ", verbose=False) -> Iterable[Tuple[str, str]]:
    """
    Iterating over the words file is slightly trickier than you think due to 2 technicalities that are easy to forget:
        - You must strip the newline at the end;
        - You need to specify a sep=" ", because although Python splits on spaces by default, it uses a special
          algorithm to do so (https://stackoverflow.com/a/30271689/9352077) that drops some Unicode.
          Try " 898".split().

    Hence, we abstract it. The result is completely in TEXT form, even if the second part is a number.
    """
    for stripped_line in iterateTxt(open_file_handle, verbose=verbose):
        parts = stripped_line.split(sep=sep)
        if len(parts) > 1:
            yield sep.join(parts[0:-1]), parts[-1]


def saveWordsFile(counts: Counter, out_path: Path):
    with open(out_path, "w", encoding="utf-8") as handle:
        for word, count in sorted(counts.items(), key=lambda x: x[1], reverse=True):
            handle.write(word + " " + str(count) + "\n")


def wordsFileToCounter(words_path: Path, sep=" ") -> Counter:
    c = Counter()
    with open(words_path, "r", encoding="utf-8") as handle:
        for word, count in iterateWordsFile(handle, " "):
            if word not in c:  # Necessary (but not sufficient) to fix a bug where whitespace was left inside words. This condition catches that newlines were added to words, causing something like "a\nb 69" to show up as an ignored line "a" and a line "b 69" which would overwrite any count for b earlier in the file.
                c[word] = int(count)
    return c


@timeit
def mergeWordFiles(word_files: List[Path], out_file: Path, delete_afterwards: bool=False,
                   trim_hapax_every: int=100000):
    """
    :param trim_hapax_every: To mitigate against very large tails, trim words of count 1 (hapax legomena)
                             every this many files.
    TODO: You could implement an extra "safety valve" that detects when the dictionary goes over a certain size and
          then starts trimming off the tail (count = 1, then 2, then 3, ...) until the size is under the threshold again.
    """
    # Collect
    total_counter = Counter()
    for idx, word_file in enumerate(word_files):
        wprint(f"\nReading word file {word_file.name}...")
        new_counts = wordsFileToCounter(word_file)

        wprint("Adding counts...")
        for word, count in tqdm(new_counts.items(), total=len(new_counts)):
            total_counter[word] += count

        wprint("Size of total counter:", intsep(len(total_counter)))
        if (idx+1) % trim_hapax_every == 0:
            print("\tTrimming...")
            for word in list(total_counter.keys()):  # A list of all keys is better than a copy of the dictionary.
                if total_counter[word] == 1:
                    del total_counter[word]  # I've read that .pop() doesn't allow garbage collection the same way.
            gc.collect()  # We can't rely on the interpreter to decide when to garbage-collect those del'd items.
            print("\tAfter trimming hapax legomena:", intsep(len(total_counter)))

    # Save
    saveWordsFile(total_counter, out_file)

    # Delete
    if delete_afterwards:
        for word_file in word_files:
            word_file.unlink()


def fixWhiteSpace(words_path: Path, overwrite=True):
    """
    Due to a bug (that has now been fixed), the corpus was being split on
    only spaces rather than tabs, newlines, ... and hence the latter all
    ended up in words. The newlines caused the most havoc, but the others
    are also wrong.

    For example: the string
    '''
    Hello world, how
    are you doing? Like, how
    are you?
    '''
    supposedly contains the word "how\nare" twice, which then shows up
    in the word file as
    '''
    how
    are 2
    '''
    which then overwrote the old count for "are" and also lost 2 "how"s.

    This function overwrites a word file with correct counts. Note that a space
    can never be present in a word, which means that we can differentiate between
    numbers that are words and numbers that are counts:
    '''
    efg 1
    abc 69420
    abd 1
    '''
    could come from a word "efg" and a word "abc\t69420\nabd". There's no space in front of the 69420
    so it isn't a count written by the word file.

    You can't solve it by only allowing counts lower than the previous count, because
    then the 0 in the following example will invalidate all words below it:
    '''
    Trendprodukte	cbd 1
    schmerzmittel	0
    Essential 1
    '''
    Note that the word here is "schmerzmittel\t0\nEssential".
    """
    stack = []
    actual_counts = Counter()
    with open(words_path, "r", encoding="utf-8") as handle:
        for line in iterateTxt(handle):
            parts = line.split(" ")  # Should be 1 or 2 parts since space is only written by the saver.
            stack.extend(parts[0].split())
            if len(parts) > 1:  # The last part is a number.
                assert len(parts) == 2
                assert parts[-1].isnumeric()  # Assign this count to everything in the stack.
                count = int(parts[-1])
                for word in stack:
                    if word:
                        actual_counts[word] += count
                stack = []

    saveWordsFile(actual_counts, words_path.with_stem(words_path.stem + "_fixed") if not overwrite else words_path)


@timeit
def trimWordFile(words_path: Path, minimum: int) -> Path:
    """
    Removes all words with count < minimum.
    For OSCAR, setting minimum = 10 eliminates about 80% of all words to iterate over, greatly speeding up BPE.
    """
    removed = 0
    new_path = generatePathForTrimmedFile(words_path)
    with open(new_path, "w", encoding="utf-8") as out_handle:
        with open(words_path, "r", encoding="utf-8") as in_handle:
            for w,c in iterateWordsFile(in_handle):
                if int(c) >= minimum:
                    out_handle.write(f"{w} {c}\n")
                else:
                    removed += 1

    print("Removed", removed, "words from the count file.")
    return new_path


@timeit
def cleanWordFile(words_path: Path) -> Path:
    """
    Here's how the cleaning system works:
        - Filter out words that are too long or weirdly numerical.
        - Clean away several characters (e.g. control characters) inside the remaining words.
        - Normalise some character like hyphen-esque characters.
        - Filter out words whose characters don't match a limited set of characters.
            - Using a given set of regexes, this filtering step is attempted to be explained. (Even if no explanation is found, you STILL filter it out.)

    About length filtering: for the Dutch OSCAR corpus, there seem to be erroneously long entries, as well as non-Dutch entries.
    However, there are also long entries that are legit and longer than some non-legit entries:
        - geschillenbeslechtingsprocessen: 31 chars
        - socialehuisvestingsmaatschappijen: 33 chars
        - beroepsaansprakelijkheidsverzekeraars: 37 chars
        - toezichthoudersaansprakelijkheidsverzekering: 44 chars

    There are more entries of 45+ chars than VS Code can count, but I haven't found any legit one. You can scroll
    through them with the regex:
        \S{45}\S* [0-9]
    or if you want to eliminate Unicode and numbers:
        [A-z]{45}[A-z]* [0-9]

    There is also a non-alphanumeric line the tokenizer just got stuck on, consisting of the character "怨" 21000 times.
    In short: we need some cleaning before processing.

    Some watertight rules:
        - Delete any word that's way too long.
        - Delete any word that contains an underscore.
        - Delete any word that has disjunct numbers in it (highly likely to be an internet username; will exclude "H2O2" and "H2CO3" and so on, but not CO2 or H2O)

    I'm also quite tempted to lowercase all-caps words. For non-acronyms, this is a stylistic and not a linguistic effect
    and the word means the same thing in uppercase or lowercase. There are also a bunch of incorrect camelcased words in
    the corpus, but since there exist honest camelcased words (iPod, kWh, mAh, pH, dB ...) I will let those slide.
    I'm hoping that the sheer size of OSCAR means that true words are orders of magnitude more impactful than exceptions.

    Most of the spam is not at frequency 1, but rather 10-100. At frequency 1 you find authentic words:
        appeltiramisu 1
        pluimveevertegenwoordigers 1
        arbeidstherapeuthische 1
        nachtkompas 1
        karaokeboter 1
        smartlappenlijst 1
        kroegentochtdeelnemer 1
        boektrailerfenomeen 1
        Beeldenstormprogramma 1
        frikandellenpartjes 1
        Tegenlichtserie 1
        uitreikingsbrief 1
        Bedrijfsservicedesk 1
        tussenlening 1
        kortingfout 1
        dameszwemteam 1
    """
    MAX_LENGTH = 60
    CONTAINS_DISJUNCT_DIGITS = re.compile(r"[A-z]+[0-9]+[A-z]+[0-9]+[A-z0-9]*")
    # CAPITAL_CHARACTERS = re.compile(r"[A-Z]")

    new_path = generatePathForCleanedFile(words_path)
    removed = 0
    error_explanations = Counter()
    with open(new_path, "w", encoding="utf-8") as out_handle:
        with open(words_path, "r", encoding="utf-8") as in_handle:
            for raw_word, raw_count in iterateWordsFile(in_handle):
                # Long input
                if len(raw_word) > MAX_LENGTH:
                    removed += 1
                    continue

                # Internet usernames
                if "_" in raw_word or CONTAINS_DISJUNCT_DIGITS.search(raw_word):
                    removed += 1
                    continue
                # if raw_word[0].islower() and CAPITAL_CHARACTERS.search(raw_word[1:]):
                #     continue

                # Sanitise word
                raw_word = dashes_to_replace.sub("-", raw_word)
                raw_word = CONTROL_CHARACTERS.sub("", FILTERED_WHITESPACE.sub("", C_AND_R.sub("", raw_word)))

                # Check for alphabet violation
                for c in raw_word:
                    if not all_allowed_characters.match(c):
                        # Violation found; this word will be filtered out. Try to find explanation.
                        explanation = None
                        for key, pattern in UNICODE_EXPLANATIONS.items():
                            if pattern.match(c):
                                explanation = key
                                break

                        error_explanations[explanation or "unexplained"] += 1
                        removed += 1
                        break
                else:  # Never reached 'break' so you never violated.
                    out_handle.write(f"{raw_word} {raw_count}\n")


    print("Removed", removed, "words from the count file.")
    print("Distribution:", error_explanations)
    print("Unexplained:", error_explanations["unexplained"])
    return new_path


# def detectForeignAlphabets(string: str):
#     for p in UNICODE_EXPLANATIONS.values():
#         match = p.search(string)
#         if match:
#             return match
#     return None


ACCENTS = re.compile("[áàäâãéèëêíìïîóòöôõúùüû]")


def recomposeAccents(word_file: Path):
    """
    Unicode normalisation with NFD means that accented characters are decomposed into two characters (the letter and
    the accent). NFC reverts that.
    """
    from tokenizers.normalizers import NFC
    n = NFC()

    counts = wordsFileToCounter(word_file)
    new_counts = Counter()
    for word, count in counts:
        new_counts[n.normalize_str(word)] += count

    saveWordsFile(new_counts, word_file)


def reaccentWordFile(word_file: Path, lemmata_with_accents: Iterable[str]):
    """
    Uses lemmata which presumably come from a corpus WITH accents, to put accents into the word file at those lemmata.
    """
    import tokenizers.normalizers as tn
    normalizer = tn.Sequence([tn.NFD(), tn.StripAccents()])  # Need the "D" because it "D"ecomposes an accented letter into the letter and its accent, e.g. ä -> a¨ (in Unicode, that looks like: ä)

    accent_map = dict()
    for lemma in lemmata_with_accents:
        if ACCENTS.search(lemma):
            accent_map[normalizer.normalize_str(lemma)] = lemma
    print(accent_map)

    out_path = generatePathForAccentedFile(word_file)
    with open(out_path, "w", encoding="utf-8") as out_handle:
        with open(word_file, "r", encoding="utf-8") as in_handle:
            for word, count in iterateWordsFile(in_handle):
                accented = accent_map.get(word, word)
                out_handle.write(accented + " " + count + "\n")

    return out_path


def generatePathForCleanedFile(path: Path):
    return path.with_stem(path.stem + "_cleaned")


def generatePathForTrimmedFile(path: Path):
    return path.with_stem(path.stem + "_trimmed")


def generatePathForAccentedFile(path: Path):
    return path.with_stem(path.stem + "_reaccented")


def wordfileToBpeCorpus(wordfile: Path, do_pretokenise=False) -> Iterable[str]:
    """
    Converts a file
        apple 5
        banana-pear 3
        ...
    to sentences
        Ġapple Ġapple Ġapple Ġapple Ġapple
        Ġbanana - pear Ġbanana - pear Ġbanana - pear

    or, if pretokenisation is turned on (do NOT do this when using HuggingFace's ByteLevel pretokeniser already!),
        apple apple apple apple apple
        banana-pear banana-pear banana-pear
    """
    from .hf_corpora import punctuation_regex_str
    PUNCTUATION_PATTERN = re.compile("(" + punctuation_regex_str + ")")  # Could've just been ([\-]+) but the other punctuation could also prove useful for wordfiles where you forgot to split off periods etc.
    LARGEST_STRING_LEN  = 1_000

    with open(wordfile, "r", encoding="utf-8") as handle:
        for word, count in iterateWordsFile(handle):
            if do_pretokenise:
                word = " ".join(PUNCTUATION_PATTERN.split("Ġ" + word))  # Add spaces around (groups of) punctuation. Note that the pattern's capture group ( ) keeps the separator. https://stackoverflow.com/a/2136580/9352077
            count = int(count)  # Note: you can't just multiply the string you want to return by this count, because you'll run into memory errors (kinda hard to reserve 400 million bytes of stack space).
            max_at_once = max(1, LARGEST_STRING_LEN // (len(word)+1))
            while count > 0:
                new_count = max(0, count - max_at_once)
                diff      = count - new_count
                count -= diff
                yield diff*(" " + word)


def intersectLexiconCounts(all_lemmata_wordfile: Path,
                           subset_lexicon: Iterable[str], subset_name: str) -> Optional[Counter]:
    """
    Get the intersection between the morphological lexicon and the word count lexicon.

    Why would you not use the counts from the morphological lexicon (if it has them in the first place)?
        - Frequencies in e-Lex are often 0, and the max (for "de" and "en") is 250k.
        - Frequencies in OSCAR have max ~250M, which is 1000x more information. According to Zipf's law, all counts should have
          increased proportionally, meaning their relative contribution is the same (~ 1/rank), so any weighting done with
          a larger corpus shouldn't skew towards the higher frequencies.

    Here's what we could do:
        1. Collect all surface forms for a lemma in e-Lex that has morphology.
        2. Use OSCAR's cleaned frequencies to assign those counts.
        3. Sum the counts per lemma and store that as lemma weights.
        4. Recalculate the above metric using the frequencies.

    An easier way, purely matching on lemmas:
        1. Find lemma in OSCAR.
        2. Use that frequency.
    Note that this approach neglects all verb conjugations and all plural nouns.
    """
    if all_lemmata_wordfile is None:  # Impossible to identify which cache file it would be.
        return None

    cache_path = makeDownloadPath() / f"{all_lemmata_wordfile.stem} ⊗ {subset_name}.txt"  # Path depends on the two files it intersects, otherwise it would be used even if you switched languages.
    if not cache_path.exists():
        if not all_lemmata_wordfile.exists():  # Impossible to fill the cache.
            return None

        counter = Counter()

        # Collect subset
        for w in subset_lexicon:
            counter[w] = 1  # We effectively add the lexicon to the corpus.

        # Look up their counts
        with open(all_lemmata_wordfile, "r", encoding="utf-8") as handle:
            for word, count in iterateWordsFile(handle):
                if word in counter:
                    counter[word] += int(count)

        # Cache these filtered counts
        with open(cache_path, "w", encoding="utf-8") as handle:
            for word, count in counter.items():
                handle.write(f"{word} {count}\n")

    return wordsFileToCounter(cache_path)


def loadAndWeightLexicon(all_lemmata_wordfile: Path,
                         subset_lexicon: Iterable[str], subset_name: str,
                         reweighting_function: Callable[[float],float]) -> Dict[str, float]:
    """
    Takes care of converting word counts (integers) to weights (floats)
    and returns a queriable object even if no counts exist.
    """
    lemma_weights = intersectLexiconCounts(all_lemmata_wordfile, subset_lexicon, subset_name)  # Fill the cache.
    if lemma_weights is None:  # Possible if there was no weights file found.
        return dict()
    else:
        lemma_weights = dict(lemma_weights)
        for word, frequency in lemma_weights.items():
            lemma_weights[word] = reweighting_function(frequency)  # Note that it's only disallowed to ADD items in an iterable, not change them.
        return lemma_weights



from .vocabulariser import *


class BPEKnockoutVocabulariser_NoTrivialKnockout(BPEKnockoutVocabulariser):
    """
    BTE without knocking out trivial merges, which are merges where all parts are quite long and thus the merges are
    likely compounds. The reason these are trivial is that if you have a subword like 'lighthouse', it likely came from a merge

        light house -> lighthouse

    and because the word "lighthouse" would contain the one occurrence of this merge, BPE-knockout can just undo that last
    merge to memorise the correct answer. But this is not so interesting, because this knockout affects no other words,
    and furthermore, perhaps it isn't a bad thing for a language model to have separate subwords for light, house and
    lighthouse.

    So, to isolate how much of the performance is gained just from memorising these trivial examples, we don't knock
    them out and then see how much performance gain still applies.

    (I show in my thesis that knockout's performance gain is indeed more than this.)
    """
    def __init__(self, initial_tokeniser: BPEArtifacts, config: BTEConfig, long_part_threshold: int, holdout: Holdout, iteration_evaluator: IntermediateEvaluator, quiet: bool=True):
        super().__init__(initial_tokeniser=initial_tokeniser, config=config, holdout=holdout, iteration_evaluator=iteration_evaluator, quiet=quiet)
        self._long_part_threshold = long_part_threshold

    def _rankOldMergesForKnockout(self, tk: BTE, reference: ModestDataset, blame_tuples_once: bool=False) -> list[MergeBlame]:
        filtered_results = super()._rankOldMergesForKnockout(tk, reference, blame_tuples_once)
        return [m for m in filtered_results if not m.merge.isTrivial(minimum=self._long_part_threshold)]

    def _getMetadata(self, tk: BTE, reference: ModestDataset) -> dict:
        d = super()._getMetadata(tk, reference)
        d["identifiers"]["tokeniser"] += "_keeptrivial"
        return d


class BPEngineer:
    """
    Wrapper that runs all kinds of diagnostics on a BTE tokeniser.

    These methods are kept out of the tokeniser class because we don't want to hide them, but they do pollute the
    interface of the main tokeniser.
    """

    def __init__(self, bte: BTE):
        self.bte = bte

    def findInvariantViolations(self) -> List[str]:
        """
        There is an invariant (proved in my thesis under the name "original functional sin", OFS) in vanilla BPE that
        says that every type in the vocabulary can be formed by exactly one merge, because (as the proof shows) there
        are no more strings in the corpus from which BPE can learn a second merge for that type after the first has been learnt.

        With reification, this invariant can be broken. TODO: This is only superficially true. You can add any merge you want to the tokeniser, but only one will be applied per type.
        This function gives you the types for which the invariant is broken.
        """
        multimerge_types = []
        for typ, merges in self.bte.merge_graph.merges_of.items():
            if len(merges) > 1:
                # self.bte._print(f"Found type with multiple merges: '{typ}' formed by {' and '.join(['<' + '+'.join(merge.parts) + '>' for merge in merges])}")
                multimerge_types.append(typ)
        return multimerge_types

    def findDisabledTypes(self) -> List[str]:
        """
        It is easy to prove that a necessary condition for a type in the vocabulary of a BPE tokeniser to ever be formed
        is that the tokeniser forms it when you tokenise the type itself in string form:
            - In strings where it doesn't appear, it can never be formed.
            - In strings where it does appear, the only thing that could change vs. when it is alone as a string is that
              a neighbouring character could merge with one of the type's characters, but then it can once again no longer
              be formed as a token because now you always have an extra character you can't lose.

        Because vanilla BPE's segmentation is an exact replay of its vocabularisation, we also know that in vanilla BPE
        a type is actually formed when it is offered to the tokeniser as a string.

        With knockout, this is no longer true. E.g.: the type "_bruids" is formed by a merge "_bru + ids" which becomes
        "_bru + id + s" after knockout of the merge "id + s". However, because the merge "_bru + id" already exists and
        appears before this merge, there will never be a token sequence [_bru, id, s] when you arrive at the triplet merge.
        At best, there will be sequences [_bruid, s], for which you don't know a merge.

        Reification can correct such disabilities, but likely also create more of them.
        """
        unformable_types = []
        for typ in self.bte.merge_graph.vocab:
            tokens = self.bte.tokenise(typ)
            if len(tokens) > 1:
                # self.bte._print(f"Found disabled type: '{typ}' -> {tokens}")
                unformable_types.append(typ)
        return unformable_types

    def findLeafChains(self):
        """
        Goal: Find chains of merges whose result only appear in a single merge, and specifically those
              that then feed into a leaf.
        TODO:
            What we don't detect is chains that are distantly connected. Imagine you have 5 merges.
            M1's type is only used in M2. M2's type is used in M3 and M4. M3's type is only used in M5. That looks like
                    /--- M3 --- M5
            M1 --- M2
                    \--- M4
            wherein M1-M2 and M3-M5 are two separate chains.
        """
        singletons_in_order = [merge for merge in self.bte.merge_graph.merges if len(self.bte.merge_graph.merges_with[merge.childType()]) == 1]
        singletons = {merge.childType(): merge for merge in singletons_in_order}
        # print(singletons)
        # print(len(singletons))

        chains = []
        exist_in_chains = set()
        for merge in reversed(singletons_in_order):  # Due to the order, you know that the largest chain to which a type could possibly belong will have been found by the time you get to it.
            if merge.childType() in exist_in_chains:
                continue

            chain = []
            while True:
                chain.append(merge)
                for part in merge.parts:  # Try to go DOWN the chain.
                    merge = singletons.get(part)
                    if merge is not None:  # Early exit when you find a type.
                        break
                else:  # If there was no type found (no early exit), the chain is done.
                    break
            exist_in_chains.update([merge.childType() for merge in chain])
            chains.append(chain[::-1])

        class Chain:
            def __init__(self, chain: List[Merge]):
                self.as_list: List[Merge] = chain
                self.connected_to_leaves: List[Merge] = []

        # chains.sort(key=lambda chain: len(chain))
        # lprint(chains)

        # Now find the chains that end in a leaf; these are purpose-built chains for that leaf,
        # or alternatively a compound built from two words that would've been standalone words
        # but ended up as part of an even bigger standalone word and hence became single-use instead of a leaf.
        chains = {chain[-1].childType(): Chain(chain) for chain in chains}
        leaves = [merge for merge in self.bte.merge_graph.merges if len(self.bte.merge_graph.merges_with[merge.childType()]) == 0]
        for leaf in leaves:
            for part in leaf.parts:
                chain = chains.get(part)
                if chain is not None:
                    chain.connected_to_leaves.append(leaf)

        lprint(
            map(lambda chain: (chain.connected_to_leaves[0].childType(), chain.as_list + chain.connected_to_leaves),  # what is displayed
                sorted(  # sort them first by chain length and then by leaf type length
                    filter(  # get all chains that end in 1 leaf
                        lambda chain: len(chain.connected_to_leaves) == 1,
                        chains.values()
                    ),
                    key=lambda chain: (len(chain.as_list), len(chain.connected_to_leaves[0].childType()))
                )
            )
        )

    def printSurroundingGraph(self, t: str):
        """
        Print the vertices (types) that emanate from the given type, and its siblings (i.e. types emanating from its
        parents), along with the relevant edges (merges).
        """
        involved_merges = self.bte.merge_graph.merges_with[t]

        children = [m.childType() for m in involved_merges]
        spouses = set()
        for child in children:
            spouses.update(self.bte.merge_graph.merges_of[child][0].parts)
        if children:
            spouses.remove(t)

        parent_merge = self.bte.merge_graph.merges_of[t][0]
        parents = parent_merge.parts

        siblings = set()
        parental_spouses = set()
        for parent in parents:
            children_of_parent = self.bte.merge_graph.merges_with[parent]
            for child_merge in children_of_parent:
                parental_spouses.update(child_merge.parts)
                siblings.add(child_merge.childType())
        siblings.remove(t)

        print("Things it makes:", len(children), children)
        print("Things it combines with:", len(spouses), spouses)
        print("Things its parents produce:", len(siblings), siblings)
        print("Parents and their spouses:", len(parental_spouses), parental_spouses)

    def buildTransitionGraph(self, pretoken: str):
        def bufferToBits(buffer: str):
            bits = []
            i = 2
            while i < len(buffer)-1:
                if buffer[i] == " ":
                    bits.append("1")
                    i += 2  # Next will definitely be a character. Jump over that character to the next possible space.
                else:
                    bits.append("0")
                    i += 1

            return "".join(bits)

        buffer = " " + " ".join(self.bte._initialTokens(pretoken)) + " "
        state  = BpeMergeState(bufferToBits(buffer))

        starting_bitstring = state.bitstring

        frontier: list[tuple[str,BpeMergeState]] = [(buffer,state)]  # Buffers as in ._finalTokens, along with the reference where to store the children.
        found: dict[str, BpeMergeState]          = {state.bitstring: state}  # Bit strings that uniquely identify each segmentation that has been formed, without storing tokens. Saves up to 2x characters versus storing buffers.
        while frontier:
            buffer, state = frontier.pop(0)
            types = buffer[1:-1].split(" ")
            types.pop()  # Last type won't ever start a merge, so it's useless to check merges for it.

            index_in_buffer = 0  # Start on the first space in the buffer.
            for t in types:
                for m in self.bte.merges_starting_with.get(t, []):
                    # Is this a valid merge at this point?
                    if buffer[index_in_buffer:index_in_buffer+len(m[1])] == m[1]:
                        # Yes, so we know what the new buffer looks like, what its bitstring is, and that it's a child of the current buffer.
                        # If we haven't seen it before, add it to the frontier.
                        new_buffer    = buffer[:index_in_buffer] + m[2] + buffer[index_in_buffer+len(m[1]):]
                        new_bitstring = bufferToBits(new_buffer)
                        if new_bitstring not in found:
                            new_state = BpeMergeState(new_bitstring)
                            found[new_bitstring] = new_state
                            frontier.append((new_buffer,new_state))
                        else:
                            new_state = found[new_bitstring]

                        state.children.append(new_state)

                index_in_buffer += len(t) + 1  # Move to the next space in the buffer.

        return found[starting_bitstring]


class BpeMergeState:

    def __init__(self, bitstring: str):
        self.bitstring = bitstring
        self.children: List[BpeMergeState] = []

    def __eq__(self, other) -> bool:
        return self.bitstring == other.bitstring

    def __hash__(self) -> int:
        return hash(self.bitstring)

    def __repr__(self, indent: int=0, already_printed: set=None) -> str:
        if already_printed is None:
            already_printed = set()

        result = "|   "*indent + self.bitstring
        if self in already_printed:
            result += " (^)\n"
        else:
            result += "\n"
            already_printed.add(self)
            for state in self.children:
                result += state.__repr__(indent+1, already_printed)

        return result



from transformers import AutoTokenizer, PreTrainedTokenizerBase
from huggingface_hub import hf_hub_download

from tktkt.interfaces.identifiers import AutoVocabSpecs, AutoVocab, repairAbsoluteSpecials, areNotAbsoluteSpecials, \
    SpecialsExtended
from tktkt.models.huggingface.wrapper import HuggingFacePreprocessorForWords

from .vocabulariser import *
from .tokeniser import *


class AutoKnockout:
    """
    Normally, the way to apply BPE-knockout is to apply the Vocabulariser to a given tokeniser, which gives a .json file.
    This file can then be loaded into a BTE tokeniser in another runtime.

    Alternatively, you apply the Vocabulariser AND immediately load the result into a BTE object in the same runtime.
    That's what this class does.
    (There's a good case to be made that these should be different methods on the Vocabulariser, except the result of
    a Vocabulariser has never been an object and always a Path.)
    """

    def __init__(self, config: BTEConfig):
        self.config = config

    def from_pretrained(self, checkpoint: str, original_specials: AutoVocabSpecs[WithSpecials], reference: ModestDataset) -> BTE[WithSpecials]:
        tkz: PreTrainedTokenizerBase = AutoTokenizer.from_pretrained(checkpoint)
        return self.from_objects(
            preprocessor=HuggingFacePreprocessorForWords(tkz),
            vocab=AutoVocab.fromTokenizer(tkz, original_specials),
            merges=AutoMerges.from_tokenizer(tkz),
            reference=reference
        )

    def from_objects(self, preprocessor: Preprocessor, vocab: Vocab[WithSpecials], merges: MergeList, reference: ModestDataset) -> BTE[WithSpecials]:
        artifacts = CacheableBPEArtifacts(types=list(vocab), merges=merges)
        artifacts.setPreprocessors(preprocessor)
        return self.from_artifacts(
            artifacts=artifacts,
            original_specials=SpecialsExtended(vocab.specials, vocab.UNK),
            reference=reference
        )

    def from_artifacts(self, artifacts: BPEArtifacts, original_specials: SpecialsExtended[WithSpecials], reference: ModestDataset) -> BTE[WithSpecials]:
        new_artifacts = BPEKnockoutVocabulariser(initial_tokeniser=artifacts, config=self.config).vocabulariseFromModest(reference=reference)

        if not areNotAbsoluteSpecials(original_specials.specials):  # not not absolute == absolute. These will need correction after knockout!
            types = new_artifacts.getVocabulary()
            specials, unk_id = repairAbsoluteSpecials(count(types), original_specials.specials, original_specials.unk)
        else:
            specials, unk_id = original_specials.specials, original_specials.unk

        return BTE(
            new_artifacts.preprocessorEffective(),
            vocab=new_artifacts.getVocabulary(SpecialsExtended(specials, unk_id)),
            merges=new_artifacts.getMerges(),
            metadata=self.config
        )


class AutoMerges:

    @staticmethod
    def from_pretrained(checkpoint: str) -> MergeList:
        """
        Automatically constructs the file path, and ALSO imputes the tokeniser file by getting it from the
        HuggingFace tokeniser with the given name.
        """
        try:
            cache = Path(hf_hub_download(repo_id=checkpoint, filename="tokenizer.json"))
            uses_json = True
        except:
            try:
                cache = Path(hf_hub_download(repo_id=checkpoint, filename="merges.txt"))
                uses_json = False
            except:
                raise RuntimeError(f"Could not find (or access) the online tokeniser file for HuggingFace model '{checkpoint}'.")

        if uses_json:
            with open(cache, "r", encoding="utf-8") as handle:
                merges = json.load(handle)["model"]["merges"]
        else:
            with open(cache, "r", encoding="utf-8") as handle:
                merges = [line.strip() for line in handle if not line.startswith("#version")]

        assert isinstance(merges, list)
        return merges

    @staticmethod
    def from_tokenizer(tokenizer: PreTrainedTokenizerBase) -> MergeList:
        """
        Only works for HuggingFace models that were NOT loaded from a path, but just using a name
        (the former have an empty name, that's why).
        """
        name = tokenizer.name_or_path
        if not name:
            raise ValueError("Model has no name!")

        return AutoMerges.from_pretrained(name)



"""
Wrapper around an iterator that dynamically (and reproducibly!) separates it into
examples that are allowed to pass through and examples that aren't, with a certain
percentage.
In other words: it does holdout. You can select whether you want either or both of
the splits to be let through.

Given the same deterministic iterator, this class will always select the same
examples to assign to the train or test split, even across multiple object instances.
"""
import numpy.random as npr
from typing import TypeVar, Iterable

T = TypeVar("T")


class Holdout:

    def __init__(self, train_fraction: float, seed: int=0):
        self.threshold = train_fraction
        self.seed = seed

    def __call__(self, iterator: Iterable[T], train: bool=False, test: bool=False) -> Iterable[T]:
        if self.threshold == 0.0 or self.threshold == 1.0:  # In these degenerate cases, just have the whole dataset available.
            train = True
            test = True

        self.it = iterator
        self.rng = npr.default_rng(seed=self.seed)

        for output in self.it:  # This lets the iterator yield
            train_split = self.rng.random() < self.threshold
            if (train_split and train) or (not train_split and test):  # This skips some of the iterator's results.
                yield output



import dataclasses
from enum import Enum

from modest.interfaces.morphologies import MorphologyVisitor, FreeMorphSplit, MorphSplit


class ReferenceMode(str, Enum):  # The str parent allows JSON serialisation: https://stackoverflow.com/a/51976841/9352077
    NONE      = 1
    MORPHEMIC = 2
    ONLY_FREE_MORPHS = 3

    def toMethod(self) -> MorphologyVisitor:
        if   self == ReferenceMode.ONLY_FREE_MORPHS:
            return FreeMorphSplit()
        elif self == ReferenceMode.MORPHEMIC:
            return MorphSplit()
        else:
            raise NotImplementedError()

    def toLetter(self) -> str:
        if   self == ReferenceMode.ONLY_FREE_MORPHS:
            return "f"
        elif self == ReferenceMode.MORPHEMIC:
            return "m"
        elif self == ReferenceMode.NONE:
            return ""
        else:
            raise NotImplementedError()


class ReifyMode(str, Enum):
    """
    Chooses between enabling the following reification features:
        - Fixing diverging triplets created by knockout;
        - Turning triplets back into binary merges by linking them to existing merges;
        - Turning triplets back into binary merges by creating new merges, when linking is not possible.

    There is no setting for creating new merges without linking existing merges (just "MAKE"), because realistically, nobody wants this.
    """
    NONE                  = 1
    LINK                  = 2
    LINK_AND_MAKE         = 3
    FIX                   = 4
    FIX_AND_LINK          = 5
    FIX_AND_LINK_AND_MAKE = 6

    NONE_CASCADE = 7  # Alters BPE-knockout to cascade its merges rather than rewiring them into tuple merges. No fixing, linking, or making can thus be done.

    def does_nothing(self):
        return self in {ReifyMode.NONE, ReifyMode.NONE_CASCADE}

    def does_fix(self):
        return self in {ReifyMode.FIX, ReifyMode.FIX_AND_LINK, ReifyMode.FIX_AND_LINK_AND_MAKE}

    def does_link(self):
        return self in {ReifyMode.LINK, ReifyMode.LINK_AND_MAKE, ReifyMode.FIX_AND_LINK, ReifyMode.FIX_AND_LINK_AND_MAKE}

    def is_backwards_compatible(self):
        """Whether new types will NOT be added to the vocabulary by reification. Equivalent to 'does_not_make()'."""
        return self not in {ReifyMode.LINK_AND_MAKE, ReifyMode.FIX_AND_LINK_AND_MAKE}


class AnnealingTime(str, Enum):
    BEFORE = 1
    AFTER  = 2
    BOTH   = 3

    def toLetter(self) -> str:
        if   self == AnnealingTime.BEFORE:
            return "before"
        elif self == AnnealingTime.AFTER:
            return "after"
        elif self == AnnealingTime.BOTH:
            return "surround"
        else:
            raise NotImplementedError()


@dataclasses.dataclass
class KnockoutConfig:
    reference: ReferenceMode = ReferenceMode.NONE
    min_vocab_size: int = 0
    relative_blame_minimum: float = 0.5
    blame_tuples_once: bool = False  # When computing blame on a merge with more than 2 tokens, each instance can either be seen as one application (True) or as the amount of spaces that are concatenated by it (False). For example: if a merge (a,b,c,d) takes place, 'True' counts it as 1 application, whilst 'False' counts as 3 applications.


@dataclasses.dataclass
class AnnealingConfig:
    reference: ReferenceMode = ReferenceMode.NONE
    when: AnnealingTime = AnnealingTime.BEFORE  # Only matters when annealing is set to something other than None.
    max_vocab_size: int = 1_000_000
    absolute_application_minimum: int = 3       # A merge has to be applied without issue at least this many times to be worth adding.
    relative_amenability_minimum: float = 0.80  # A merge has to be applied without issue at least with this ratio.


@dataclasses.dataclass
class BTEConfig:
    knockout:   KnockoutConfig = dataclasses.field(default_factory=KnockoutConfig)
    annealing: AnnealingConfig = dataclasses.field(default_factory=AnnealingConfig)
    reify: ReifyMode = ReifyMode.NONE
    iterations: int = 1

    # Legacy arguments that are not really relevant anymore.
    weighted_training: bool = False  # Shown in the paper to not really matter.



__version__ = "2026.01.01"

from .model.config import BTEConfig, KnockoutConfig, AnnealingConfig, ReferenceMode, ReifyMode, AnnealingTime
from .model.tokeniser import BTE
from .model.vocabulariser import BPEKnockoutVocabulariser
from .model.auto import AutoKnockout, AutoMerges
from .util.project.config import Pℛ𝒪𝒥ℰ𝒞𝒯, lexiconWeights



from typing import Union
from dataclasses import dataclass
from enum import Enum

from tqdm.auto import tqdm

from tktkt.interfaces import Vocab
from tktkt.util.printing import warn

__all__ = ["MergeGraph", "Merge", "MergeExplanation", "MergeAsTuple", "MergeList"]


MergeAsTuple = tuple[int, str, str]  # (priority, "a b c", "abc")
MergeOnDisk = Union[str, list[str], tuple[str,...]]  # "a b c" or ("a", "b", "c") with implicit priority.
MergeList   = list[MergeOnDisk]


class MergeExplanation(Enum):
    PREEXISTING     = 1
    KNOCKOUT        = 2
    ANNEALED        = 3
    REPAIRED        = 4
    REIFIED         = 5
    ALREADY_REIFIED = 6


@dataclass
class Merge:
    priority: int
    parts: list[str]
    explanation: MergeExplanation

    def __lt__(self, other):
        return self.priority < other.priority

    def __hash__(self):
        return self.asTuple().__hash__()

    def asTuple(self) -> MergeAsTuple:
        """
        Returns a 3-tuple of the merge's priority, the string of what its parts
        look like when separated by spaces, and the string of what they look like
        joined together. Both of the latter are padded by spaces.
        """
        return (
            self.priority,
            " " + " ".join(self.parts) + " ",
            " " + self.childType() + " "
        )

    def childType(self) -> str:
        return "".join(self.parts)

    def isTrivial(self, minimum: int) -> bool:
        """
        A merge is trivial if all its parts are at least as long as a given number.
        This indicates that the merge is just making a giant compound, which is, trivially, over-eager.
        """
        return all([len(part) >= minimum for part in self.parts])


class MergeGraph:
    """
    Handles the relationships between BPE types and merges.

    Has 4 data structures:
        - self.vocab: the vertices representing the types in the vocabulary.
        - self.merges: list of merges in order, as objects, each storing their priority and the list of their merged types.
        - self.merges_with: dictionary from type to a list of references to merge objects whose list contains that type.
        - self.merges_of: dictionary from type to the list of references to merge objects whose parts concatenate to form that type.
                          In vanilla BPE or BPE with just knockout, this list always has length 1 due to original functional sin.
    """

    def __init__(self, vocab: Vocab, raw_merges: MergeList, quiet=True):
        self.next_merge = 0  # == 1 + max([m.priority for m in self.merges]), not always len(self.merges) due to knockout.

        # Initialise graph
        self.merges: list[Merge] = []
        self.vocab: Vocab = vocab
        self.merges_with: dict[str, list[Merge]] = dict()
        self.merges_of:   dict[str, list[Merge]] = dict()  # Note: in deterministic BPE vocabularisers, only one merge is learnt per type. And in deterministic BPE tokenisers, only one merge is ever used to form each type. Thus, in all practical scenarios, this list contains only one object (but it could contain more for BPE-dropout, in theory).

        # Fill graph
        for raw_type in tqdm(vocab, desc="ADDING VERTICES", disable=quiet):
            self.addVertex(raw_type)

        for raw_merge in tqdm(raw_merges, desc="LINKING VERTICES", disable=quiet):
            self.addArc(raw_merge)

    def addVertex(self, type_to_add: str):
        if " " in type_to_add:
            raise ValueError(f"The type '{type_to_add}' contains a space. This is illegal.")

        if type_to_add not in self.vocab:
            self.vocab.add(type_to_add)
        if type_to_add not in self.merges_with:
            self.merges_with[type_to_add] = []
        if type_to_add not in self.merges_of:
            self.merges_of[type_to_add]   = []

    def addArc(self, merge_to_add: MergeOnDisk, add_missing_atoms: bool=False) -> Merge:
        """
        Adds arcs to the merge graph, and the resulting type if necessary.
        Also returns the constructed merge object for diagnostic purposes.

        :param merge_to_add: tupled or space-separated merge, e.g. "ab cd e".
        """
        parts = self._parseRawMerge(merge_to_add, check_vocab=not add_missing_atoms)
        if add_missing_atoms:
            for part in parts:
                if part not in self.vocab:
                    if len(part) > 1:
                        warn(f"Adding atom '{part}' with more than 1 character to the vocabulary.")
                    self.addVertex(part)

        new_merge = Merge(self.next_merge, parts, explanation=MergeExplanation.PREEXISTING)
        new_type = new_merge.childType()

        if new_type not in self.vocab:
            self.addVertex(new_type)
        self.merges.append(new_merge)
        for part in set(parts):  # set() in case there is a duplicate part.
            self.merges_with[part].append(new_merge)
        self.merges_of[new_type].append(new_merge)

        self.next_merge += 1
        return new_merge

    def knockout(self, type_to_delete: str):
        """New implementation that is equivalent to that in the paper, but more modularised."""
        # Collect all the information we have about this type.
        affected_merges   = self.merges_with[type_to_delete]
        replacement_parts = self.merges_of[type_to_delete][0].parts

        # Rewire all the affected merges.
        for m in list(affected_merges):  # list() because the whole point of this loop is to shrink the list.
            self.rewire(
                m.childType(),
                        (" " + "  ".join(m.parts)          + " ")  # Two spaces because " xy xy ".replace(" xy ", " x y ") == " x y xy " due to the middle space not being allowed to be the last space of one match and the first of another.
                .replace(" " + type_to_delete              + " ",
                         " " + " ".join(replacement_parts) + " ")
                .replace("  ", " ")
                .strip()
            )
            m.explanation = MergeExplanation.KNOCKOUT

        # Cut the type out of the graph.
        self.cascade(type_to_delete, cleanup=False)

    def rewire(self, type_to_rewire: str, new_merge: MergeOnDisk) -> Merge:
        if type_to_rewire not in self.vocab:
            raise ValueError(f"Type does not exist: {type_to_rewire}")
        # print("Rewiring", type_to_rewire, "to", new_merge)

        merge = self.merges_of[type_to_rewire][0]
        old_parts = merge.parts

        # Unlink all the old parts
        for part in set(old_parts):
            self.merges_with[part].remove(merge)

        # Link up new parts
        new_parts = self._parseRawMerge(new_merge)
        merge.parts = new_parts
        for part in set(new_parts):
            self.merges_with[part].append(merge)

        return merge

    def cascade(self, type_to_delete: str, cleanup: bool=True) -> set[str]:
        """
        Applies knockout to the entire tree of descendants of the given type.
        This is easy to do: just prevent the type from being formed again, and all its descendants are blocked too.

        :param cleanup: If false, the blocked merges and their resulting types keep existing in the tokeniser, but will
                        just never be formed again.
        """
        if type_to_delete not in self.vocab:
            raise ValueError(f"Type does not exist: {type_to_delete}")

        if self.inAlphabet(type_to_delete):
            warn(f"Type {type_to_delete} is in the alphabet. Knockout will result in some inputs being impossible to represent.")

        # Detach this root from the rest of the graph
        for m in self.merges_of[type_to_delete]:
            for parent in set(m.parts):
                self.merges_with[parent].remove(m)

        # Handle the cascade
        if not cleanup:
            types_to_delete = {type_to_delete}
        else:  # You might think that cascaded knockout can be done recursively, but it's more difficult than that since the BPE merge graph is a DAG, not a tree.
            frontier        = {type_to_delete}
            types_to_delete = set()
            while frontier:
                current_type = frontier.pop()
                types_to_delete.add(current_type)

                affected_types = set()
                for m in self.merges_with[current_type]:
                    affected_types |= set(m.parts)
                frontier |= affected_types - types_to_delete

        for type_to_delete in types_to_delete:
            # Remove from vocab.
            self.vocab.pop(type_to_delete)

            # Remove the merge that made this.
            for m in self.merges_of[type_to_delete]:
                self.merges.remove(m)

            # Forget that you had merges for this.
            self.merges_of.pop(type_to_delete)
            self.merges_with.pop(type_to_delete)

        return types_to_delete

    def _parseRawMerge(self, merge_on_disk: MergeOnDisk, check_vocab: bool=True) -> list[str]:
        parts = merge_on_disk.split(" ") if isinstance(merge_on_disk, str) else list(merge_on_disk)
        if check_vocab and not all([p in self.vocab for p in parts]):
            raise ValueError(f"The merge '{merge_on_disk}' contains types not in the vocab yet.")
        if any([p == "" for p in parts]):
            raise ValueError(f"The merge '{merge_on_disk}' seems to have double spaces.")
        return parts

    def getRawMerges(self) -> list[str]:
        return [" ".join(merge.parts) for merge in sorted(self.merges)]  # Have to sort explicitly because priorities aren't returned, and they are sometimes changed during execution causing the list to be out of order.

    def getPaddedMerges(self) -> list[MergeAsTuple]:
        return [merge.asTuple() for merge in self.merges]

    def inAlphabet(self, typ: str) -> bool:
        return len(self.merges_of[typ]) == 0



#!/usr/bin/env python

"""
Source: https://github.com/amazon-science/statistical-byte-pair-encoding
License:
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
Copyright (c) 2015 University of Edinburgh

SPDX-License-Identifier: MIT-0

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import argparse
import codecs
import re
import sys
import math

from typing import List, Iterable, TextIO
from collections import defaultdict
from tqdm.auto import tqdm

from tktkt.interfaces.preprocessors import Preprocessor
from tktkt.factories.preprocessors import SennrichSpaceMarker, IdentityMapper, PretokeniserSequence, AddWordBoundary, HyphenMode, IsolatePunctuation, OnWhitespace
from tktkt.preparation.boundaries import BoundaryMarker

from .heap import HeapWithInverseIndex

def error(message: str):
    sys.stderr.write(message)

DEFAULT_PREPROCESSOR = Preprocessor(  # [Bauwens] This is a minimal preprocessor that is correct, but there are better ones (to support numbers and so on). E.g.: "BPE-knockout is (very) cool." -> ["BPE", "-", "knockout</w>", "is</w>", "(</w>", "very</w>", ")</w>", "cool</w>", ".</w>"]
    IdentityMapper(),
    IdentityMapper(),
    PretokeniserSequence([
        IsolatePunctuation(HyphenMode.EXCLUDED, protect_apostrophes_without_spaces=False),
        OnWhitespace(destructive=True),
        AddWordBoundary(SennrichSpaceMarker),
        IsolatePunctuation(HyphenMode.ONLY),
    ])
)


class VocabCounter:
    """
    Extract the word-level vocabulary from a file
    (which can be a plain text or a dictionary file (i.e. (words, freq) pairs).
    """

    def __init__(self, input_handles: Iterable[Iterable[str]], dictionary_mode: bool=False,
                 marker: BoundaryMarker=SennrichSpaceMarker, preprocessor: Preprocessor=DEFAULT_PREPROCESSOR):
        self.forward_index = dict()  # tuple of word characters -> token ID
        self.inverse_index = []      # token ID (list index) -> tuple of word characters
        self.counter = []            # token ID (list index) -> count in corpus
        self.marker = marker
        self.preprocessor = preprocessor

        for file_handle in input_handles:
            for line in tqdm(file_handle):  # Although it should be a handle, any iterator that produces strings works.
                if dictionary_mode:
                    fields = line.strip("\r\n ").split(" ")
                    assert len(fields) == 2  # Each line in the file looks like e.g. "potato 69420".
                    pretokens = self.preprocessor.do(fields[0])
                    count     = int(fields[1])
                else:
                    pretokens = self.preprocessor.do(line.strip("\r\n "))  # [Bauwens] Support for byte-level conversion, punctuation splitting, etc...
                    count     = 1

                for p in pretokens:
                    self._add_pretoken(p, count=count)

    def _add_pretoken(self, pretoken: str, count=1):
        if not pretoken:
            return

        # [Bauwens] Proper treatment of boundary markers.
        pretoken_as_tuple = tuple(self.marker.atomise(pretoken))

        # The original code
        index = self.forward_index.get(pretoken_as_tuple)
        if index is None:
            index = len(self.inverse_index)
            self.forward_index[pretoken_as_tuple] = index
            self.inverse_index.append(pretoken)
            self.counter.append(0)
        self.counter[index] += count

    def substitute(self, old_word, new_word):
        pos = self.forward_index[old_word]
        del self.forward_index[old_word]
        self.forward_index[new_word] = pos
        self.inverse_index[pos] = new_word

    def get_word(self, pos):
        return self.inverse_index[pos]

    def get_counts_from_index(self, pos):
        return self.counter[pos]

    def __len__(self):
        return len(self.counter)

    class Iterator:
        def __init__(self, vocab):
            self.vocab = vocab
            self.pos = 0

        def __next__(self):
            if self.pos < len(self.vocab.counter):
                rv = (self.pos, self.vocab.inverse_index[self.pos], self.vocab.counter[self.pos])
                self.pos += 1
                return rv
            else:
                raise StopIteration

    def __iter__(self):
        return self.Iterator(self)


class PairStats:
    """
    Class for handling the pairs of operations. It is basically a wrapper
    around a heap of operations, with additional functions for updating counts
    when a new operation has been selected.
    """
    def __init__(self, vocab, probabilistic):
        self.vocab = vocab
        self.probabilistic = probabilistic

        raw_stats = defaultdict(int)  # From pairs to counts
        self.vocab_entries_for_pair = defaultdict(set)  # From pairs to lists of indices in vocab
        for pos, word, freq in tqdm(self.vocab, total=len(self.vocab)):
            word_pair_stats = self.get_pair_stats_from_word(word)
            for pair, count in word_pair_stats.items():
                raw_stats[pair] += count * freq
                self.vocab_entries_for_pair[pair].add(pos)

        # For probabilistic BPE we need the counts of the produced items
        if self.probabilistic:
            self.produced_count = defaultdict(int)
            self.n_running_symbols = 0
            # Store the counts of the initial units (characters)
            for _, word, freq in tqdm(self.vocab, total=len(self.vocab)):
                for unit in word:
                    self.produced_count[unit] += freq
                    self.n_running_symbols += freq

        # stats_heap will contain pairs (freq, word)
        self.stats_heap = HeapWithInverseIndex(value_score_function=self._get_scoring_function(),
                                               use_score_caching=True,
                                               key_function=lambda x: x[1])
        for pair in tqdm([(i[1], i[0]) for i in raw_stats.items()], desc="HEAP", total=len(raw_stats)):
            self.stats_heap.insert(pair)

    def get_pair_stats_from_word(self, word, filter_elems=None):
        """
        Computes the statistics for pairs in a word. If filter_elems is given,
        only pairs involving these elements are returned.

        Note that there is a mismatch between standard and probabilistic BPE.
        Consider as an example the word '10002'. The standard BPE
        implementation reports 2 occurences of the pair '00', but when creating
        the split, the result would be '1@@ 00@@ 0@@ 2', i.e. the pair '00'
        appears only once. For standard BPE extraction we do not correct this
        in order to stay compatible with the original implementation.

        For probabilistic BPE, as we need the counts of the produced elements,
        this mismatch has to be corrected in order to avoid negative
        probabilities.
        """
        word_len = len(word)
        idx = 1
        pair_stats = defaultdict(int)
        while idx < word_len:
            prev_unit = word[idx - 1]
            unit = word[idx]
            pair = (prev_unit, unit)
            if filter_elems is None or (prev_unit in filter_elems or unit in filter_elems):
                pair_stats[pair] += 1
            if self.probabilistic:
                # If we have three consecutive equal elements, skip the next one
                if prev_unit == unit and idx < word_len - 1 and unit == word[idx + 1]:
                    idx += 2
                else:
                    idx += 1
            else:
                idx += 1
        return pair_stats

    def __len__(self):
        return len(self.stats_heap)

    def probabilistic_score_laplace(self, unit, new_unit_count):
        total_count = self.n_running_symbols
        normalization = total_count - new_unit_count
        first_count = self.produced_count[unit[0]] - new_unit_count
        second_count = self.produced_count[unit[1]] - new_unit_count
        score = new_unit_count * (math.log(new_unit_count + 1) -
                                  math.log(first_count + 1) -
                                  math.log(second_count + 1) +
                                  math.log(normalization + len(self.produced_count) + 1))
        return score

    def _get_compare_function(self):
        # We include the full tuple in the comparisons in order to break ties
        if self.probabilistic:
            return lambda a, b: (self.probabilistic_score_laplace(a[1], a[0]), a[1]) > \
                (self.probabilistic_score_laplace(b[1], b[0]), b[1])
        else:
            return lambda a, b: (sum(a[0]), a[1]) > (b[0], b[1])

    def _get_scoring_function(self):
        if self.probabilistic:
            return lambda a: (self.probabilistic_score_laplace(a[1], a[0]), a[1])
        else:
            return lambda a: a

    def pop_max(self):
        """
        Extract the most frequent element, create a new pair and adjust counts.
        """
        heap_max = self.stats_heap.pop_max_cached_value()
        max_elem = heap_max.value
        self.stats_heap.increase_timestep()
        freq, pair = max_elem
        first, second = pair
        pair_str = first + second

        if self.probabilistic:
            self.n_running_symbols -= freq
            self.produced_count[first] -= freq
            self.produced_count[second] -= freq
            self.produced_count[pair_str] += freq

        # This approach is taken from the original implementation. We could
        # probably optimize this processing and try to avoid jumping between
        # pairs and strings, as well as avoiding re.
        pair_str = pair_str.replace('\\', '\\\\')
        pattern = re.compile(r'(?<!\S)' + re.escape(first + ' ' + second) + r'(?!\S)')
        stats_changes = {}
        for w_index in self.vocab_entries_for_pair[pair]:
            # Update words
            old_word = self.vocab.get_word(w_index)
            new_word = " ".join(old_word)
            new_word = pattern.sub(pair_str, new_word)
            new_word = tuple(new_word.split(" "))
            self.vocab.substitute(old_word, new_word)

            freqs = self.vocab.get_counts_from_index(w_index)
            self._update_stats(pair, old_word, new_word, freqs, w_index, stats_changes)

        updated_stats = {}
        for mod_pair, freq_change in stats_changes.items():
            heap_entry = self.stats_heap.get(mod_pair)
            if not heap_entry:
                updated_stats[mod_pair] = freq_change
            else:
                updated_stats[mod_pair] = heap_entry[0] + freq_change
                self.stats_heap.invalidate_key(mod_pair)
        for mod_pair, freq in updated_stats.items():
            if freq > 0:
                self.stats_heap.insert((freq, mod_pair))

        del self.vocab_entries_for_pair[pair]
        return max_elem, heap_max.score

    def _update_stats(self, new_pair, old_word, new_word, freq, w_index, stats_changes):
        """
        Update the statistics after merging the pair new_pair.

        In this implementation we take the easy way, compute the stats for the
        old and the new word, compare them and adapt accordingly. This allows
        for an easy implementation and accomodates the slight difference
        between probabilistic and non-probabilistic counts without effort (see
        self.get_pair_stats_from_word for the mismatch).
        """
        filter_set = {new_pair[0], new_pair[1], new_pair[0]+new_pair[1]}
        old_pair_stats = self.get_pair_stats_from_word(old_word, filter_set)
        new_pair_stats = self.get_pair_stats_from_word(new_word, filter_set)
        all_pairs = set(old_pair_stats.keys()) | set(new_pair_stats.keys())
        for pair in sorted(list(all_pairs)):
            if pair == new_pair:
                continue

            freq_change = None
            if pair in new_pair_stats and pair not in old_pair_stats:
                # New pair for this word
                count = new_pair_stats[pair]
                freq_change = count * freq
                self.vocab_entries_for_pair[pair].add(w_index)
            elif pair in old_pair_stats and pair not in new_pair_stats:
                # The pair does not exist any more in this word
                count = old_pair_stats[pair]
                freq_change = -count * freq
                # The next line is conceptually correct, but triggers an
                # error that the set changed during iteration
                #~ self.vocab_entries_for_pair[pair].remove(w_index)
            else:
                # The pair is both in the new and old words
                count_diff = new_pair_stats[pair] - old_pair_stats[pair]
                if count_diff != 0:
                    freq_change = count_diff * freq

            if freq_change:
                if pair in stats_changes:
                    stats_changes[pair] += freq_change
                else:
                    stats_changes[pair] = freq_change


def adapt_num_symbols(num_symbols: int, vocab: VocabCounter, total_symbols: bool):
    """
    Handle the parameter --total_symbols.
    """
    new_num_symbols = num_symbols
    if total_symbols:
        unique_chars = set()
        for _, char_tuple, _ in vocab:
            for char in char_tuple:
                unique_chars.add(char)

        error("Number of basic characters (merges that won't be done): {0}\n".format(len(unique_chars)))
        new_num_symbols -= len(unique_chars)

    return new_num_symbols


def learn_bpe(infiles: List[Iterable[str]], outfile: TextIO,
              num_symbols_ori: int, total_symbols=False,
              probabilistic=False, frac_stopping=None, frac_stopping_average_n=100, min_frequency=2,
              is_dict=False, verbose=False,
              marker: BoundaryMarker=SennrichSpaceMarker, preprocessor: Preprocessor=DEFAULT_PREPROCESSOR):
    """
    Learn num_symbols BPE operations from vocabulary, and write to outfile.

    [Bauwens]: apparently Vilar & Federico wrote this file in such a way that it is fully compatible with the
               original BPE interface. The "probabilistic" parameter switches from BPE to S-BPE.

    :param num_symbols_ori: size of vocab OR amount of merges.
    :param total_symbols: if True, num_symbols_ori is the vocab size. If False, it is the amount of merges.
    :param probabilistic: BPE (False) or S-BPE (True). Should really be "statistical".
    :param frac_stopping: "k" parameter in the paper. Should be 0.002.
    :param frac_stopping_average_n: "M" parameter in the paper. They say 5 is okay.
    :param min_frequency: If using BPE, stops early when the argmax drops below this.
    :param is_dict: whether or not the infiles are in "dictionary mode", i.e.
        word1 count1
        word2 count2
        word3 count3 ...
    """
    # version 0.2 changes the handling of the end-of-word token ('</w>');
    # version numbering allows backward compatibility
    # [Vilar]: We should be compatible with the original 0.2 version
    outfile.write('#version: 0.2\n')

    print("Generating word vocabulary...")
    vocab       = VocabCounter(infiles, is_dict, marker=marker, preprocessor=preprocessor)
    num_symbols = adapt_num_symbols(num_symbols_ori, vocab, total_symbols)

    print("Getting pair statistics...")
    pair_stats  = PairStats(vocab, probabilistic)

    ini_score = None
    frac_stopping_accum = 0.0
    num_written = 0
    for num_written in tqdm(range(num_symbols), desc="MERGES", smoothing=0.1):
        if not pair_stats:
            error('No more pairs after creating {} symbols. Stopping\n'.format(num_written))
            break
        (freq, pair), score = pair_stats.pop_max()

        if probabilistic and frac_stopping > 0.0:
            frac_stopping_accum += score[0]
            if num_written + 1 == frac_stopping_average_n:
                ini_score = frac_stopping_accum / frac_stopping_average_n
                frac_stopping_accum = 0.0
            elif (num_written + 1) % frac_stopping_average_n == 0:
                avg = frac_stopping_accum / frac_stopping_average_n
                if avg < frac_stopping * ini_score:
                    error(f'Stopping due to frac-stopping after {num_written} symbols ({avg} < {frac_stopping} * {ini_score})\n')
                    break
                else:
                    frac_stopping_accum = 0.0

        if not probabilistic and freq < min_frequency:
            error('no pair has frequency >= {0}. Stopping\n'.format(min_frequency))
            break
        if verbose:
            error('pair {0}: {1} {2} -> {1}{2} (frequency {3})\n'.format(num_written, pair[0], pair[1], freq))
        outfile.write('{0} {1}\n'.format(*pair))

    error(f"{num_written} pairs written to file.\n")
    return num_written


def create_arg_parser(subparsers=None):
    """
    Create the argument parser.

    Copied from the original implementation to be command-line compatible.
    """
    if subparsers:
        parser = subparsers.add_parser('learn-bpe',
                                       formatter_class=argparse.RawDescriptionHelpFormatter,
                                       description="learn BPE-based word segmentation")
    else:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="learn BPE-based word segmentation")

    parser.add_argument(
        '--input', '-i', type=argparse.FileType('r'), default=[sys.stdin],
        metavar='PATH', nargs="+",
        help="Input text(s) (default: standard input).")

    parser.add_argument(
        '--probabilistic', '-p', action="store_true",
        help="Use probabilistic BPE")
    parser.add_argument(
        '--frac-stopping', '-fs', type=float, default=0.0,
        help="(Probabilistic) Stop when the likelihood increase falls below this fraction of the initial one)")
    parser.add_argument(
        '--frac-stopping-average', '-fsa', type=int, default=5,
        help='"Mini-batch" size for frac-stopping computation (default: %(default)s)')
    parser.add_argument(
        '--output', '-o', type=argparse.FileType('w'), default=sys.stdout,
        metavar='PATH',
        help="Output file for BPE codes (default: standard output)")
    parser.add_argument(
        '--symbols', '-s', type=int, default=10000,
        help="Create this many new symbols (each representing a character n-gram) (default: %(default)s))")
    parser.add_argument(
        '--min-frequency', type=int, default=2, metavar='FREQ',
        help='Stop if no symbol pair has frequency >= FREQ (default: %(default)s))')
    parser.add_argument('--dict-input', action="store_true",
        help="If set, input file is interpreted as a dictionary where each line contains a word-count pair")
    parser.add_argument(
        '--total-symbols', '-t', action="store_true",
        help="Subtract number of characters from the symbols to be generated (so that '--symbols' becomes an estimate for the total number of symbols needed to encode text).")
    parser.add_argument(
        '--verbose', '-v', action="store_true",
        help="verbose mode.")

    return parser


def main():
    # Full compatibility with original implementation.
    # I do not know exactly why this is different to the standard file objects,
    # but some special utf-8 symbols are handled differently if the codecs call
    # is not present.
    sys.stderr = codecs.getwriter('UTF-8')(sys.stderr.buffer)
    sys.stdout = codecs.getwriter('UTF-8')(sys.stdout.buffer)
    sys.stdin  = codecs.getreader('UTF-8')(sys.stdin.buffer)

    arg_parser = create_arg_parser()
    args = arg_parser.parse_args()

    # Full compatibility with original implementation
    for i in range(len(args.input)):
        if args.input[i].name != '<stdin>':
            args.input[i] = codecs.open(args.input[i].name, encoding='utf-8')
    if args.output.name != '<stdout>':
        args.output = codecs.open(args.output.name, 'w', encoding='utf-8')

    learn_bpe(
        args.input, args.output, args.symbols,
        probabilistic=args.probabilistic,
        frac_stopping=args.frac_stopping,
        frac_stopping_average_n=args.frac_stopping_average,
        min_frequency=args.min_frequency,
        is_dict      =args.dict_input,
        total_symbols=args.total_symbols,
        verbose      =args.verbose,
    )


if __name__ == "__main__":
    main()


from abc import ABC, abstractmethod
from pathlib import Path

import requests
import json
from transformers import AutoTokenizer, RobertaTokenizerFast, PreTrainedTokenizerFast
from huggingface_hub import hf_hub_download

from tktkt.paths import TkTkTPaths

DEFAULT_TOKENISER_STEM = "tokenizer"


def makeDownloadPath() -> Path:
    return TkTkTPaths.pathToCheckpoints("downloads")


def AutoTokenizer_from_pretrained(path_or_name: str) -> PreTrainedTokenizerFast:
    """
    HuggingFace's AutoTokenizer.from_pretrained somehow doesn't work with local paths.
    Fine, I'll do it myself.
    """
    path = Path(path_or_name)
    if path.is_absolute():  # Get it from disk.
        return PreTrainedTokenizerFast(tokenizer_file=path.as_posix())
    else:  # Get it from the hub (or from the HF_CACHE, presumably).
        return AutoTokenizer.from_pretrained(path_or_name)


class BpeTokeniserPath(ABC):  # TODO: This is superseded by TkTkT's better interface.
    """
    Interface for representing a BPE tokeniser on disk.
    """

    def __init__(self, path: Path):
        self.path = path

    def __repr__(self):
        return self.__class__.__name__ + "(" + self.path.as_posix() + ")"

    @abstractmethod
    def exists(self) -> bool:
        pass

    @staticmethod
    @abstractmethod
    def fromName(name: str) -> "BpeTokeniserPath":
        pass

    @abstractmethod
    def loadVocabulary(self) -> dict[str,int]:
        pass

    @abstractmethod
    def loadMerges(self) -> list[str]:
        pass

    @abstractmethod
    def toFastBPE(self) -> RobertaTokenizerFast:
        pass


class SennrichTokeniserPath(BpeTokeniserPath):
    """
    Tokeniser stored as a vocab.json file and a merges.txt file in the same folder.
       - vocab.json: JSON with a single top-level dictionary, mapping each subword type string to an integer id.
       - merges.txt: Text file of BPE merges, in order of creation. Each merge is a string consisting of the merged types, separated by a space.
    """
    def __init__(self, folder: Path):
        super().__init__(folder)
        folder.mkdir(exist_ok=True, parents=True)

    def exists(self) -> bool:
        vocab, merges = self.getPaths()
        return vocab.exists() and merges.exists()

    def getPaths(self) -> tuple[Path, Path]:
        return self.path / "vocab.json", self.path / "merges.txt"

    def loadVocabulary(self) -> dict[str,int]:
        with open(self.getPaths()[0], "r", encoding="utf-8") as handle:
            return json.load(handle)

    def loadMerges(self) -> list[str]:
        with open(self.getPaths()[1], "r", encoding="utf-8") as handle:
            return [line.strip() for line in handle if not line.startswith("#version")]

    def toFastBPE(self) -> RobertaTokenizerFast:
        vocab, merges = self.getPaths()

        ###
        # Fix because HuggingFace is fake and gay
        # https://github.com/huggingface/transformers/blob/main/src/transformers/models/roberta/tokenization_roberta.py#L228
        with open(merges, "r", encoding="utf-8") as handle:
            lines = handle.readlines()
        if not lines[0].startswith("#version"):
            lines.insert(0, "#version-knockout\n")
            with open(merges, "w", encoding="utf-8") as handle:
                handle.writelines(lines)
        ###

        return RobertaTokenizerFast(vocab_file=vocab.as_posix(), merges_file=merges.as_posix())  # Will apply byte-based pretokeniser.

    # @staticmethod
    # def fromName(name: str) -> BpeTokeniserPath:
    #     return SennrichTokeniserPath(PATH_MODELBASE / name)


class HuggingFaceTokeniserPath(BpeTokeniserPath):
    """
    Tokeniser as stored by HuggingFace's 'tokenizers' library as a single JSON file.
    """
    def __init__(self, json_path: Path):
        super().__init__(json_path)
        json_path.parent.mkdir(exist_ok=True, parents=True)

    def exists(self) -> bool:
        return self.path.exists()

    def getAsdict(self):
        with open(self.path, "r", encoding="utf-8") as handle:
            return json.load(handle)

    def loadVocabulary(self) -> dict[str,int]:
        return self.getAsdict().get("model", dict()).get("vocab", dict())

    def loadMerges(self) -> list[str]:
        return self.getAsdict().get("model", dict()).get("merges", [])

    def toFastBPE(self) -> RobertaTokenizerFast:
        return AutoTokenizer_from_pretrained(self.path.as_posix())

    @staticmethod
    def fromName(name: str, use_hf_cache: bool=True) -> BpeTokeniserPath:
        """
        Automatically constructs the file path, and ALSO imputes the tokeniser file by getting it from the
        HuggingFace tokeniser with the given name.
        (The reason you can't do this in getAsdict() is because the name isn't known there, only the path.)
        """
        PATH_DATA_TEMP = makeDownloadPath()
        if use_hf_cache:
            try:
                cache = Path(hf_hub_download(repo_id=name, filename="tokenizer.json"))
                return HuggingFaceTokeniserPath(cache)
            except:
                try:
                    path_vocab  = Path(hf_hub_download(repo_id=name, filename="vocab.json"))
                    path_merges = Path(hf_hub_download(repo_id=name, filename="merges.txt"))
                    return SennrichTokeniserPath(path_vocab.parent)
                except:
                    raise RuntimeError(f"Could not find (or access) the online tokeniser file for HuggingFace model '{name}'.")
        else:
            temp_cache = HuggingFaceTokeniserPath(PATH_DATA_TEMP / name.replace("/", "--") / f"{DEFAULT_TOKENISER_STEM}.json")
            if not temp_cache.exists():
                try:
                    Path(hf_hub_download(repo_id=name, filename="tokenizer.json", local_dir=temp_cache.path.parent))
                    # fetchAndCachedict(f"https://huggingface.co/{name}/raw/main/tokenizer.json",
                    #                   cache_folder=cache.path.parent, stem=DEFAULT_TOKENISER_STEM)
                except:  # Likely means that this is a GPT2 tokeniser where there was no tokenizer.json and instead there is only a vocab and merge file.
                    temp_cache = SennrichTokeniserPath(temp_cache.path.parent)
                    if not temp_cache.exists():
                        try:
                            path_vocab  = Path(hf_hub_download(repo_id=name, filename="vocab.json", local_dir=temp_cache.path))
                            path_merges = Path(hf_hub_download(repo_id=name, filename="merges.txt", local_dir=temp_cache.path))
                        except:
                            raise RuntimeError(f"Could not find (or access) the online tokeniser file for HuggingFace model '{name}'.")

                        # Reformat the local files.
                        with open(path_vocab, "r", encoding="utf-8") as handle:
                            vocab = json.load(handle)
                        with open(path_merges, "r", encoding="utf-8") as handle:
                            merges = handle.readlines()
                            merges = merges[merges[0].startswith("#version"):]

                        vocab_path, merges_path = temp_cache.getPaths()
                        with open(vocab_path, "w", encoding="utf-8") as handle:
                            json.dump(vocab, handle, ensure_ascii=False, indent=4)
                        with open(merges_path, "w", encoding="utf-8") as handle:
                            handle.write("\n".join(merges))

            return temp_cache

    @staticmethod
    def fromTokeniser(tk_model: RobertaTokenizerFast) -> "BpeTokeniserPath":
        """
        Only works for HuggingFace models that were NOT loaded from a path, but just using a name
        (the former have an empty name, that's why).
        """
        name = tk_model.name_or_path
        if not name:
            raise ValueError("Model has no name!")

        return HuggingFaceTokeniserPath.fromName(name)


def fetchAndCacheDict(url: str, stem: str, cache_folder: Path=None) -> dict:
    """
    Download something with json syntax from the internet,
    store it locally, and return it as a dictionary.
    If it already exists locally, it is not downloaded again, to protect against outages.
    """
    if cache_folder is None:
        cache_folder = makeDownloadPath()

    path = cache_folder / (stem + ".json")
    if path.exists():
        with open(path, "r", encoding="utf-8") as handle:
            j = json.load(handle)
    else:
        response = requests.get(url)
        try:
            j = response.json()  # Convert response to JSON dict.
        except:
            raise RuntimeError(f"Could not retrieve JSON file (status code: {response.status_code}) from URL: {url}")

        cache_folder.mkdir(parents=True, exist_ok=True)
        with open(path, "w", encoding="utf-8") as handle:
            json.dump(j, handle, ensure_ascii=False, indent=4)

    return j



"""
Global configuration of the project.
The purpose of this script is to provide all paths to necessary files,
and to impute those files if they don't exist.

Sits at the bottom of the import hierarchy: it should import the bare minimum
of project files, and be imported itself by most project files.

It itself cannot run anything. Imputation should only be done by calling its
functions.
"""
from abc import abstractmethod, ABC
from typing import Callable, Optional, Iterable, Dict
from langcodes import Language
from dataclasses import dataclass
from pathlib import Path

import json
import math
import langcodes

from modest.interfaces.datasets import ModestDataset
from modest.interfaces.morphologies import WordDecompositionWithFreeSegmentation

# None of the below files import the config.
from bpe_knockout.util.storage import BpeTokeniserPath, SennrichTokeniserPath
from bpe_knockout.util.datahandlers.wordfiles import loadAndWeightLexicon


class ImputablePath(ABC):

    def __init__(self, path: Path):
        self.path = path

    def exists(self):
        return self.path.exists()

    @abstractmethod
    def impute(self, language: Language):
        pass


LINEAR_WEIGHTER  = lambda f: f
ZIPFIAN_WEIGHTER = lambda f: 1 + math.log10(f)

@dataclass
class ProjectConfig:
    # Name of the tested language, e.g. "English". Should exist, so that its standardised language code can be looked up.
    language_name: str
    # MoDeST dataset that contains morphological segmentations, both bound and free.
    morphologies: Optional[ModestDataset[WordDecompositionWithFreeSegmentation]]
    # Text file that contains the frequencies of words in a large corpus. Each line is a word, a space, and an integer.
    lemma_counts: Optional[ImputablePath]
    # File(s) for constructing the base tokeniser (see above).
    base_tokeniser: BpeTokeniserPath
    # Function to run over the frequencies to turn them into the weights that are used later on.
    reweighter: Callable[[float], float] = LINEAR_WEIGHTER

    def langTag(self) -> str:
        return langcodes.find(self.language_name).to_tag()

    def imputeLemmaCounts(self):
        if self.lemma_counts is not None and not self.lemma_counts.exists():
            self.lemma_counts.impute(langcodes.find(self.language_name))

    def imputeMorphologies(self):  # This is technically no longer relevant since MoDeST auto-imputes on a .generate() call.
        if self.morphologies is None:
            raise RuntimeError("Cannot impute morphologies because no path was given.")

        if not self.morphologies._rerouted:
            self.morphologies._files()

    def imputeTokeniser(self):
        from ..datahandlers.bpetrainer import BPETrainer

        # TODO: These should probably go somewhere else.
        VOCAB_SIZE = 40_000
        BYTE_BASED = True

        if not self.base_tokeniser.exists():
            language_object = langcodes.find(self.language_name)

            # Before training from scratch, try to impute just the vocab if you already have the merges.
            retrain_from_scratch = True
            if isinstance(self.base_tokeniser, SennrichTokeniserPath):
                base_vocab, base_merges = self.base_tokeniser.getPaths()
                if base_merges.exists() and not base_vocab.exists():
                    retrain_from_scratch = False

                    vocab = BPETrainer.deduceVocabFromMerges(base_merges, byte_based=True)  # assume it's byte-based
                    with open(base_vocab, "w", encoding="utf-8") as handle:
                        json.dump(vocab, handle, ensure_ascii=False, indent=4)

            # If you couldn't deduce anything, write vocab and merges.
            if retrain_from_scratch:
                print(f"{language_object.display_name()} tokeniser not found. Training...")
                self.imputeLemmaCounts()

                trainer = BPETrainer(vocab_size=VOCAB_SIZE, byte_based=BYTE_BASED)
                trainer.train_hf(wordfile=self.lemma_counts.path,  # Imputed above.
                                 out_folder=self.base_tokeniser.path)  # Takes about 3h40m (English).


#########################################################################################


class OscarWordFile(ImputablePath):

    def impute(self, language: Language):
        print(f"{language.display_name()} lemma weights not found. Counting...")
        from ..datahandlers.hf_corpora import dataloaderToCounts, generateDataloader_Oscar, punctuationPretokeniserExceptHyphens
        dataloader, size = generateDataloader_Oscar(langtag=language.to_tag(),
                                                    sentence_preprocessor=punctuationPretokeniserExceptHyphens(),
                                                    size_limit=30_000_000)
        counts = dataloaderToCounts(dataloader, self.path.stem)  # Takes about 28h30m (English), 4h30m (Dutch), ...
        counts.rename(self.path)


#########################################################################################


@dataclass
class Project:
    config: ProjectConfig=None

    debug_prints: bool=False
    verbosity: bool=False

Pℛ𝒪𝒥ℰ𝒞𝒯 = Project(None)
# All files access this object for paths. Because it is an object, its fields can be changed by one
# file (main.py) and those changes are available to all files. Hence, you call something like
#       PROJECT.config = setupDutch()
# in main.py.
#
# You could imagine two alternatives:
#   1. In this file, defining a CONFIG = setupDutch(). The downside is that this can only be controlled from inside this
#      file, rather than from a central control file.
#   2. In main.py, defining a CONFIG to be imported by all other files. This doesn't work because main.py is all the way
#      at the top of the import hierarchy and hence you would get circular imports.
#
# The trade-off you make by taking the current approach over approach (1) is that there are two configs defined in the
# project: one is the default above, the other is the one in main.py. The one above is used for all executions that do
# not use main.py (every instance of 'if __name__ == "__main__"' and also all executions starting from outside this
# codebase, e.g. those that import BTE), the other is used only for executing main.py.


class KnockoutDataConfiguration:

    def __init__(self, context: ProjectConfig):
        self.old_context = None
        self.new_context = context

    def __enter__(self):
        self.old_context = Pℛ𝒪𝒥ℰ𝒞𝒯.config
        Pℛ𝒪𝒥ℰ𝒞𝒯.config = self.new_context

    def __exit__(self, exc_type, exc_val, exc_tb):
        Pℛ𝒪𝒥ℰ𝒞𝒯.config = self.old_context

TemporaryContext = KnockoutDataConfiguration  # Alias for backwards compatibility with old users of the library.


def defaultTokeniserFiles() -> BpeTokeniserPath:
    Pℛ𝒪𝒥ℰ𝒞𝒯.config.imputeTokeniser()

    return Pℛ𝒪𝒥ℰ𝒞𝒯.config.base_tokeniser


def morphologyGenerator() -> Iterable[WordDecompositionWithFreeSegmentation]:
    """
    Alias for LemmaMorphology.generator that automatically uses the project's file path for morphologies.
    Without this, you would need to repeat the below statement everywhere you iterate over morphologies.
    """
    Pℛ𝒪𝒥ℰ𝒞𝒯.config.imputeMorphologies()

    return Pℛ𝒪𝒥ℰ𝒞𝒯.config.morphologies.generate()  # https://discuss.python.org/t/difference-between-return-generator-vs-yield-from-generator/2997


def lexiconWeights(override_reweighter: Callable[[float],float]=None) -> Dict[str, float]:
    """
    Alias for loadAndWeightLexicon that automatically uses the project's word file, morphologies, and reweighting function.
    """
    Pℛ𝒪𝒥ℰ𝒞𝒯.config.imputeLemmaCounts()

    return loadAndWeightLexicon(
        all_lemmata_wordfile=Pℛ𝒪𝒥ℰ𝒞𝒯.config.lemma_counts.path,  # Imputed above.  TODO: Possibly doesn't work when lemma_counts is None though.
        subset_lexicon=(obj.word for obj in morphologyGenerator()),  # The generator will impute morphologies itself.
        subset_name=Pℛ𝒪𝒥ℰ𝒞𝒯.config.morphologies.card().name + "-" + Pℛ𝒪𝒥ℰ𝒞𝒯.config.langTag(),
        reweighting_function=Pℛ𝒪𝒥ℰ𝒞𝒯.config.reweighter if override_reweighter is None else override_reweighter
    )
